
class Meta(type):
    ALL = []

    def __getitem__(self, arg):
        return self.ALL[arg]

    def __len__(self):
        return len(self.ALL)


class TagsCategory(metaclass=Meta):

    tag_name = None
    ALL = []
    ranked = True


    @classmethod
    def higher(cls, tag):
        if not cls.ranked:
            raise Exception('This category is not ranked')
        index = cls.ALL.index(tag)
        if index == len(cls.ALL) - 1:
            return tag
        return cls.ALL[index + 1]

    @classmethod
    def lower(cls, tag):
        if not cls.ranked:
            raise Exception('This category is not ranked')

        index = cls.ALL.index(tag)
        if index == 0:
            return tag
        return cls.ALL[index - 1]

    @classmethod
    def add(cls, tag, value):
        index = cls.ALL.index(tag)
        new_value = index + value
        if new_value < 0:
            return cls.ALL[0]
        if new_value >= len(cls.ALL):
            return cls.ALL[-1]
        return cls.ALL[new_value]

    @property
    def nb_tags(self):
        return len(Tags.ALL)

    @staticmethod
    def get_tags_names():
        return Tags.ALL

    @classmethod
    def __getitem__(cls, item):
        return Tags.ALL[item]

    @classmethod
    def add_value(cls, tag_manager, value):
        from maidi import TagManager
        from copy import deepcopy
        if not isinstance(tag_manager, TagManager):
            tag_manager = TagManager(tag_manager)

        new_array = deepcopy(tag_manager.tags)
        for idx_track, track in enumerate(new_array):
            for idx_bar, bar in enumerate(track):
                for idx_value, tag in enumerate(bar):
                    if tag.startswith(cls.tag_name):
                        new_array[idx_track][idx_bar][idx_value] = cls.add(tag, value)

        return TagManager(new_array)


    @classmethod
    def replace_tag(cls, tag_manager, tag_to_change, new_tag):
        from maidi import TagManager
        if not isinstance(tag_manager, TagManager):
            tag_manager = TagManager(tag_manager)

        return tag_manager.replace_tag(tag_to_change, new_tag)

    @classmethod
    def replace_all_with_tag(cls, tag_manager, new_tag):
        from maidi import TagManager
        from copy import deepcopy
        if not isinstance(tag_manager, TagManager):
            tag_manager = TagManager(tag_manager)

        new_array = deepcopy(tag_manager.tags)
        for idx_track, track in enumerate(new_array):
            for idx_bar, bar in enumerate(track):
                for idx_value, tag in enumerate(bar):
                    if tag.startswith(cls.tag_name):
                        new_array[idx_track][idx_bar][idx_value] = new_tag

        return TagManager(new_array)




class Density(TagsCategory):

    tag_name = 'CONTROL_DENSITY__'
    lowest = 'CONTROL_DENSITY__LOWEST'
    lower = 'CONTROL_DENSITY__LOWER'
    low = 'CONTROL_DENSITY__LOW'
    medium_low = 'CONTROL_DENSITY__MEDIUM_LOW'
    medium = 'CONTROL_DENSITY__MEDIUM'
    high = 'CONTROL_DENSITY__HIGH'
    very_high = 'CONTROL_DENSITY__VERY_HIGH'
    ALL = [lowest, lower, low, medium_low, medium, high, very_high]

class MinRegister(TagsCategory):
    tag_name = 'CONTROL_MIN_REGISTER__'

    lowest = 'CONTROL_MIN_REGISTER__lowest'
    contrabass = 'CONTROL_MIN_REGISTER__contrabass'
    bass = 'CONTROL_MIN_REGISTER__bass'
    baritone = 'CONTROL_MIN_REGISTER__baritone'
    tenor = 'CONTROL_MIN_REGISTER__tenor'
    contralto = 'CONTROL_MIN_REGISTER__contralto'
    alto = 'CONTROL_MIN_REGISTER__alto'
    mezzo_soprano = 'CONTROL_MIN_REGISTER__mezzo_soprano'
    soprano = 'CONTROL_MIN_REGISTER__soprano'
    sopranissimo = 'CONTROL_MIN_REGISTER__sopranissimo'
    highest = 'CONTROL_MIN_REGISTER__highest'

    ALL = [lowest, contrabass, bass, baritone, tenor, contralto, alto, mezzo_soprano, soprano, sopranissimo, highest]


class MaxRegister(TagsCategory):
    tag_name = 'CONTROL_MAX_REGISTER__'

    lowest = 'CONTROL_MAX_REGISTER__lowest'
    contrabass = 'CONTROL_MAX_REGISTER__contrabass'
    bass = 'CONTROL_MAX_REGISTER__bass'
    baritone = 'CONTROL_MAX_REGISTER__baritone'
    tenor = 'CONTROL_MAX_REGISTER__tenor'
    contralto = 'CONTROL_MAX_REGISTER__contralto'
    alto = 'CONTROL_MAX_REGISTER__alto'
    mezzo_soprano = 'CONTROL_MAX_REGISTER__mezzo_soprano'
    soprano = 'CONTROL_MAX_REGISTER__soprano'
    sopranissimo = 'CONTROL_MAX_REGISTER__sopranissimo'
    highest = 'CONTROL_MAX_REGISTER__highest'
    ALL = [lowest, contrabass, bass, baritone, tenor, contralto, alto, mezzo_soprano, soprano, sopranissimo, highest]


class MinPolyphony(TagsCategory):
    tag_name = 'CONTROL_MIN_POLYPHONY__'

    zero = 'CONTROL_MIN_POLYPHONY__0'
    one = 'CONTROL_MIN_POLYPHONY__1'
    two = 'CONTROL_MIN_POLYPHONY__2'
    three = 'CONTROL_MIN_POLYPHONY__3'
    four = 'CONTROL_MIN_POLYPHONY__4'
    five = 'CONTROL_MIN_POLYPHONY__5'
    six = 'CONTROL_MIN_POLYPHONY__6'

    ALL = [zero, one, two, three, four, five, six]

class MaxPolyphony(TagsCategory):
        tag_name = 'CONTROL_MAX_POLYPHONY__'
        zero = 'CONTROL_MAX_POLYPHONY__0'
        one = 'CONTROL_MAX_POLYPHONY__1'
        two = 'CONTROL_MAX_POLYPHONY__2'
        three = 'CONTROL_MAX_POLYPHONY__3'
        four = 'CONTROL_MAX_POLYPHONY__4'
        five = 'CONTROL_MAX_POLYPHONY__5'
        six = 'CONTROL_MAX_POLYPHONY__6'

        ALL = [zero, one, two, three, four, five, six]


class SpecialNotes(TagsCategory):
    ranked = False
    tag_name = 'CONTROL_SPECIAL_NOTE__'


    eleventh = 'CONTROL_SPECIAL_NOTE__s3'
    ninth = 'CONTROL_SPECIAL_NOTE__s1'
    sixth = 'CONTROL_SPECIAL_NOTE__s5'
    diminished_ninth = 'CONTROL_SPECIAL_NOTE__h1'
    major_second = 'CONTROL_SPECIAL_NOTE__h2'
    minor_third = 'CONTROL_SPECIAL_NOTE__h3'
    major_third = 'CONTROL_SPECIAL_NOTE__h4'
    diminished_fifth = 'CONTROL_SPECIAL_NOTE__h6'
    minor_sixth = 'CONTROL_SPECIAL_NOTE__h8'
    major_sixth = 'CONTROL_SPECIAL_NOTE__h9'
    minor_seventh = 'CONTROL_SPECIAL_NOTE__h10'
    major_seventh = 'CONTROL_SPECIAL_NOTE__h11'

    ALL = [eleventh, ninth, sixth, diminished_ninth, major_second,
           minor_third, major_third, diminished_fifth, minor_sixth, major_sixth, minor_seventh, major_seventh]


class Tags:

    density = Density
    min_register = MinRegister
    max_register = MaxRegister
    min_polyphony = MinPolyphony
    max_polyphony = MaxPolyphony
    special_notes = SpecialNotes
    ALL = density.ALL + min_register.ALL + max_register.ALL + min_polyphony.ALL + max_polyphony.ALL + special_notes.ALL






DRUMS = "drums"
PIANO = "piano"
BRIGHT_PIANO = "bright_piano"
ELECTRIC_PIANO = "electric_piano"
HONKY_TONK = "honky_tonk"
ELECTRIC_PIANO_1 = "electric_piano_1"
ELECTRIC_PIANO_2 = "electric_piano_2"
HARPSICHORD = "harpsichord"
CLAVI = "clavi"
CELESTA = "celesta"
GLOCKENSPIEL = "glockenspiel"
MUSIC_BOX = "music_box"
VIBRAPHONE = "vibraphone"
MARIMBA = "marimba"
XYLOPHONE = "xylophone"
TUBULAR_BELLS = "tubular_bells"
DULCIMER = "dulcimer"
DRAWBAR_ORGAN = "drawbar_organ"
PERCUSSIVE_ORGAN = "percussive_organ"
ROCK_ORGAN = "rock_organ"
CHURCH_ORGAN = "church_organ"
REED_ORGAN = "reed_organ"
ACCORDION = "accordion"
HARMONICA = "harmonica"
TANGO_ACCORDION = "tango_accordion"
ACOUSTIC_GUITAR = "acoustic_guitar"
STEEL_GUITAR = "steel_guitar"
JAZZ_GUITAR = "jazz_guitar"
CLEAN_GUITAR = "clean_guitar"
MUTED_GUITAR = "muted_guitar"
OVERDRIVEN_GUITAR = "overdriven_guitar"
DISTORTION_GUITAR = "distortion_guitar"
HARMONIC_GUITAR = "harmonic_guitar"
ACOUSTIC_BASS = "acoustic_bass"
ELECTRIC_BASS_FINGER = "electric_bass_finger"
ELECTRIC_BASS_PICK = "electric_bass_pick"
FRETLESS_BASS = "fretless_bass"
SLAP_BASS_1 = "slap_bass_1"
SLAP_BASS_2 = "slap_bass_2"
SYNTH_BASS_1 = "synth_bass_1"
SYNTH_BASS_2 = "synth_bass_2"
VIOLIN = "violin"
VIOLA = "viola"
CELLO = "cello"
CONTRABASS = "contrabass"
TREMOLO_STRING = "tremolo_string"
PIZZICATO = "pizzicato"
HARP = "harp"
TIMPANI = "timpani"
STRING_ENSEMBLE_1 = "string_ensemble_1"
STRING_ENSEMBLE_2 = "string_ensemble_2"
SYNTH_STRING_1 = "synth_string_1"
SYNTH_STRING_2 = "synth_string_2"
CHOIR_AAHS = "choir_aahs"
CHOIR_OOHS = "choir_oohs"
SYNTH_CHOIR = "synth_choir"
ORCHESTRA_HIT = "orchestra_hit"
TRUMPET = "trumpet"
TROMBONE = "trombone"
TUBA = "tuba"
MUTED_TRUMPET = "muted_trumpet"
FRENCH_HORN = "french_horn"
BRASS_SECTION = "brass_section"
SYNTH_BRASS_1 = "synth_brass_1"
SYNTH_BRASS_2 = "synth_brass_2"
SOPRANO_SAX = "soprano_sax"
ALTO_SAX = "alto_sax"
TENOR_SAX = "tenor_sax"
BARITONE_SAX = "baritone_sax"
OBOE = "oboe"
ENGLISH_HORN = "english_horn"
BASSOON = "bassoon"
CLARINET = "clarinet"
PICCOLO = "piccolo"
FLUTE = "flute"
RECORDER = "recorder"
PAN_FLUTE = "pan_flute"
BLOWN_BOTTLE = "blown_bottle"
SHAKUHACHI = "shakuhachi"
WHISTLE = "whistle"
OCARINA = "ocarina"
SQUARE_LEAD = "square_lead"
SAWTOOTH_LEAD = "sawtooth_lead"
CALLIOPE_LEAD = "calliope_lead"
CHIFF_LEAD = "chiff_lead"
CHARANG_LEAD = "charang_lead"
VOICE_LEAD = "voice_lead"
FIFTHS_LEAD = "fifths_lead"
BASS_LEAD = "bass_lead"
PAD_NEW_AGE = "pad_new_age"
PAD_WARM = "pad_warm"
PAD_POLYSYNTH = "pad_polysynth"
PAD_CHOIR = "pad_choir"
PAD_BOWED = "pad_bowed"
PAD_METALLIC = "pad_metallic"
PAD_HALO = "pad_halo"
PAD_SWEEP = "pad_sweep"
FX_RAIN = "fx_rain"
FX_SOUNDTRACK = "fx_soundtrack"
FX_CRYSTAL = "fx_crystal"
FX_ATMOSPHERE = "fx_atmosphere"
FX_BRIGHTNESS = "fx_brightness"
FX_GLOBLINS = "fx_globlins"
FX_ECHOES = "fx_echoes"
FX_SCI_FI = "fx_sci_fi"
SITAR = "sitar"
BANJO = "banjo"
SHAMISEN = "shamisen"
KOTO = "koto"
KALIMBA = "kalimba"
BAGPIPE = "bagpipe"
FIDDLE = "fiddle"
SHANAI = "shanai"
TINKLE_BELL = "tinkle_bell"
AGOGO = "agogo"
STEEL_DRUMS = "steel_drums"
WOODBLOCK = "woodblock"
TAIKO_DRUM = "taiko_drum"
MELODIC_TOM = "melodic_tom"
SYNTH_DRUM = "synth_drum"
REVERSE_CYMBAL = "reverse_cymbal"
GUITAR_FRET_NOISE = "guitar_fret_noise"
BREATH_NOISE = "breath_noise"
SEASHORE = "seashore"
BIRD_TWEET = "bird_tweet"
TELEPHONE_RING = "telephone_ring"
HELICOPTER = "helicopter"
APPLAUSE = "applause"
GUNSHOT = "gunshot"

ALL_INSTRUMENTS = [
    DRUMS,
    PIANO,
    BRIGHT_PIANO,
    ELECTRIC_PIANO,
    HONKY_TONK,
    ELECTRIC_PIANO_1,
    ELECTRIC_PIANO_2,
    HARPSICHORD,
    CLAVI,
    CELESTA,
    GLOCKENSPIEL,
    MUSIC_BOX,
    VIBRAPHONE,
    MARIMBA,
    XYLOPHONE,
    TUBULAR_BELLS,
    DULCIMER,
    DRAWBAR_ORGAN,
    PERCUSSIVE_ORGAN,
    ROCK_ORGAN,
    CHURCH_ORGAN,
    REED_ORGAN,
    ACCORDION,
    HARMONICA,
    TANGO_ACCORDION,
    ACOUSTIC_GUITAR,
    STEEL_GUITAR,
    JAZZ_GUITAR,
    CLEAN_GUITAR,
    MUTED_GUITAR,
    OVERDRIVEN_GUITAR,
    DISTORTION_GUITAR,
    HARMONIC_GUITAR,
    ACOUSTIC_BASS,
    ELECTRIC_BASS_FINGER,
    ELECTRIC_BASS_PICK,
    FRETLESS_BASS,
    SLAP_BASS_1,
    SLAP_BASS_2,
    SYNTH_BASS_1,
    SYNTH_BASS_2,
    VIOLIN,
    VIOLA,
    CELLO,
    CONTRABASS,
    TREMOLO_STRING,
    PIZZICATO,
    HARP,
    TIMPANI,
    STRING_ENSEMBLE_1,
    STRING_ENSEMBLE_2,
    SYNTH_STRING_1,
    SYNTH_STRING_2,
    CHOIR_AAHS,
    CHOIR_OOHS,
    SYNTH_CHOIR,
    ORCHESTRA_HIT,
    TRUMPET,
    TROMBONE,
    TUBA,
    MUTED_TRUMPET,
    FRENCH_HORN,
    BRASS_SECTION,
    SYNTH_BRASS_1,
    SYNTH_BRASS_2,
    SOPRANO_SAX,
    ALTO_SAX,
    TENOR_SAX,
    BARITONE_SAX,
    OBOE,
    ENGLISH_HORN,
    BASSOON,
    CLARINET,
    PICCOLO,
    FLUTE,
    RECORDER,
    PAN_FLUTE,
    BLOWN_BOTTLE,
    SHAKUHACHI,
    WHISTLE,
    OCARINA,
    SQUARE_LEAD,
    SAWTOOTH_LEAD,
    CALLIOPE_LEAD,
    CHIFF_LEAD,
    CHARANG_LEAD,
    VOICE_LEAD,
    FIFTHS_LEAD,
    BASS_LEAD,
    PAD_NEW_AGE,
    PAD_WARM,
    PAD_POLYSYNTH,
    PAD_CHOIR,
    PAD_BOWED,
    PAD_METALLIC,
    PAD_HALO,
    PAD_SWEEP,
    FX_RAIN,
    FX_SOUNDTRACK,
    FX_CRYSTAL,
    FX_ATMOSPHERE,
    FX_BRIGHTNESS,
    FX_GLOBLINS,
    FX_ECHOES,
    FX_SCI_FI,
    SITAR,
    BANJO,
    SHAMISEN,
    KOTO,
    KALIMBA,
    BAGPIPE,
    FIDDLE,
    SHANAI,
    TINKLE_BELL,
    AGOGO,
    STEEL_DRUMS,
    WOODBLOCK,
    TAIKO_DRUM,
    MELODIC_TOM,
    SYNTH_DRUM,
    REVERSE_CYMBAL,
    GUITAR_FRET_NOISE,
    BREATH_NOISE,
    SEASHORE,
    BIRD_TWEET,
    TELEPHONE_RING,
    HELICOPTER,
    APPLAUSE,
    GUNSHOT
]




import numpy as np
import os
import pickle

TEMPLATES = [
    ("", [0, 4, 7], 1),
    ("(m3)", [0, 3, 7], 1),
    ("(m3)(b5)", [0, 3, 6], 1.0),
    ("(m3)(b5)[add6]", [0, 3, 6, 9], 1.0),
    ("(+)", [0, 4, 8], 0.8),
    ("(sus2)", [0, 2, 7], 0.85),
    ("(sus4)", [0, 5, 7], 0.85),
    ("(M7)", [0, 4, 7, 11], 0.7),
    ("(m3)(m7)", [0, 3, 7, 10], 0.9),
    ("(m3)(M7)", [0, 3, 7, 10], 0.8),
    ("(m7)", [0, 4, 7, 10], 1.0),
    ("(m3)(m7)(b5)", [0, 3, 6, 10], 0.95),
]

EXTENSIONS = [
    "",
    "7",
    "(+)",
    "(sus2)",
    "(sus4)",
]
CHORD_TYPE_TO_PITCHES = {t[0]: t[1] for t in TEMPLATES}
CHORD_TYPES = np.asarray([t[0] for t in TEMPLATES])
PITCH_CLASS = np.arange(12)

CHROMA_VECTORS = [t[1] for t in TEMPLATES]
COEFFS = np.asarray([t[2] for t in TEMPLATES])
# Convert chroma_vectors into a 12xn_chords matrix with each row being the indexes of the chroma_vector
CHROMA_VECTORS_MATRIX = np.asarray(
    [
        [(1.0 * (i % 12)) in chroma_vector for i in range(12)]
        for chroma_vector in CHROMA_VECTORS
    ]
)
CHROMA_VECTORS_MATRIX = np.concatenate(
    [np.roll(CHROMA_VECTORS_MATRIX, i, axis=1) for i in range(12)], axis=0
)
COEFFS = np.concatenate([COEFFS for i in range(12)], axis=0)

MODES = ["M", "m"]
TONALITIES = list(range(12))
ROMAN_NUMERALS = list(range(7))

SCALES = {"M": [0, 2, 4, 5, 7, 9, 11], "m": [0, 2, 3, 5, 7, 8, 11]}


this_file_path = os.path.dirname(os.path.realpath(__file__))
pitch_class_dict_path = os.path.join(this_file_path, "../metadata/pitch_classes_dict.pkl")
pitches_dict_path = os.path.join(this_file_path, "../metadata/pitches_dict.pkl")
pitches_dict_roots_path = os.path.join(
    this_file_path, "../metadata/pitches_dict_roots.pkl"
)
with open(pitch_class_dict_path, "rb") as f:
    PITCH_CLASSES_DICT = pickle.load(f)

with open(pitches_dict_path, "rb") as f:
    PITCHES_DICT = pickle.load(f)

with open(pitches_dict_roots_path, "rb") as f:
    PITCHES_DICT_ROOT = pickle.load(f)



import pytest
import numpy as np
from unittest.mock import patch, MagicMock
from maidi import MidiScore, TagManager
from maidi.integrations.api import MusicLangAPI
from maidi import instrument

API_URL = "https://api.mockmusiclang.io"
MUSICLANG_API_KEY = "testapikey"


@pytest.fixture
def setup_api():
    return MusicLangAPI(api_url=API_URL, api_key=MUSICLANG_API_KEY, verbose=True)


@pytest.fixture
def setup_score():
    instruments = [instrument.PIANO]
    score = MidiScore.from_empty(instruments=instruments, nb_bars=8, ts=(4, 4), tempo=120)
    return score

@pytest.fixture
def setup_score_two_tracks():
    instruments = [instrument.PIANO, instrument.VIOLIN]
    score = MidiScore.from_empty(instruments=instruments, nb_bars=8, ts=(4, 4), tempo=120)
    return score


def test_determine_bars_step(setup_api):
    api = setup_api

    step = api._determine_bars_step(nb_added_bars_step=4, remaining_bars=10)
    assert step == 4

    step = api._determine_bars_step(nb_added_bars_step=None, remaining_bars=10)
    assert step == min(api.MAX_CONTEXT, 10)


def test_create_mask(setup_api, setup_score):
    api = setup_api
    score = setup_score
    score = score.add_silence_bars(8)
    mask = api._create_mask(score, 4)

    assert mask.shape == (score.nb_tracks, score.nb_bars)
    assert np.array_equal(mask[:, -4:], np.ones((score.nb_tracks, 4)))
    assert np.array_equal(mask[:, :-4], np.zeros((score.nb_tracks, score.nb_bars - 4)))



@patch.object(MusicLangAPI, 'predict')
def test_extend(mock_predict, setup_api, setup_score):
    api = setup_api
    score = setup_score

    def predict_side_effect(predicted_score, *args, **kwargs):
        return predicted_score  # Return the input segment as the result

    mock_predict.side_effect = predict_side_effect

    extended_score = api.extend(score, nb_bars_added=4, model="control_masking_large", nb_added_bars_step=None,
                                chords=None, tags=None, timeout=120, temperature=0.95, polling_interval=1)

    assert mock_predict.called
    assert extended_score is not None
    assert extended_score.nb_bars == score.nb_bars + 4

    # Testing with nb_added_bars_step
    extended_score = api.extend(score, nb_bars_added=8, model="control_masking_large", nb_added_bars_step=4,
                                chords=None, tags=None, timeout=120, temperature=0.95, polling_interval=1)

    assert mock_predict.called
    assert extended_score is not None
    assert extended_score.nb_bars == score.nb_bars + 8

@patch.object(MusicLangAPI, 'predict')
def test_extend_wrong_number_of_tags_track(mock_predict, setup_api, setup_score_two_tracks):
    api = setup_api
    score = setup_score_two_tracks

    def predict_side_effect(predicted_score, *args, **kwargs):
        return predicted_score

    mock_predict.side_effect = predict_side_effect
    mask, tags, chords = score.get_empty_controls()
    tags = TagManager(tags)
    tags_wrong = tags[1:, :4]  # NB BARS GOOD = 4, but wrong number of tracks

    with pytest.raises(ValueError) as e:
        api.extend(score, nb_bars_added=4, model="control_masking_large", nb_added_bars_step=None,
                                chords=None, tags=tags_wrong, timeout=120, temperature=0.95, polling_interval=1)

    assert "Wrong number of tracks in tags" in str(e.value)

@patch.object(MusicLangAPI, 'predict')
def test_extend_wrong_number_of_bars(mock_predict, setup_api, setup_score):
    api = setup_api
    score = setup_score

    def predict_side_effect(predicted_score, *args, **kwargs):
        return predicted_score

    mock_predict.side_effect = predict_side_effect
    mask, tags, chords = score.get_empty_controls()
    tags = TagManager(tags)
    tags_wrong = tags[:, :2]

    with pytest.raises(ValueError) as e:
        api.extend(score, nb_bars_added=4, model="control_masking_large", nb_added_bars_step=None,
                                chords=None, tags=tags_wrong, timeout=120, temperature=0.95, polling_interval=1)

    assert "Wrong number of bars in tags" in str(e.value)

@patch.object(MusicLangAPI, 'predict')
def test_extend_wrong_number_of_chords(mock_predict, setup_api, setup_score):
    api = setup_api
    score = setup_score

    mask, tags, chords = score.get_empty_controls()
    from maidi import ChordManager
    chords_wrong = ChordManager(chords[:2])

    with pytest.raises(ValueError) as e:
        api.extend(score, nb_bars_added=4, model="control_masking_large", nb_added_bars_step=None,
                                chords=chords_wrong, tags=None, timeout=120, temperature=0.95, polling_interval=1)

    assert "Wrong number of chords" in str(e.value)


if __name__ == "__main__":
    pytest.main()



import pytest
import numpy as np
from unittest.mock import patch, MagicMock
from maidi import MidiScore
from maidi import instrument
import os

@pytest.fixture
def setup_score():
    instruments = [instrument.PIANO]
    score = MidiScore.from_empty(instruments=instruments, nb_bars=8, ts=(4, 4), tempo=120)
    return score

def test_dilate_time_single_factor(setup_score):
    score = setup_score
    factor = 2.0
    original_times = [track["time"].copy() for track in score.tracks[score.track_keys[0]]]
    original_durations = [track["duration"].copy() for track in score.tracks[score.track_keys[0]]]

    dilated_score = score.dilate_time(factor)

    for bar in range(len(dilated_score.bars)):
        track = dilated_score.tracks[dilated_score.track_keys[0]][bar]
        assert np.array_equal(track["time"], (original_times[bar] * factor).astype(np.int32))
        assert np.array_equal(track["duration"], (original_durations[bar] * factor).astype(np.int32))

def test_dilate_time_multiple_factors(setup_score):
    score = setup_score
    factors = [1.0, 1.5, 2.0, 0.5, 1.0, 1.5, 2.0, 0.5]
    original_times = [track["time"].copy() for track in score.tracks[score.track_keys[0]]]
    original_durations = [track["duration"].copy() for track in score.tracks[score.track_keys[0]]]

    dilated_score = score.dilate_time(factors)

    for bar in range(len(dilated_score.bars)):
        track = dilated_score.tracks[dilated_score.track_keys[0]][bar]
        factor = factors[bar]
        assert np.array_equal(track["time"], (original_times[bar] * factor).astype(np.int32))
        assert np.array_equal(track["duration"], (original_durations[bar] * factor).astype(np.int32))

def test_change_ts_no_dilate(setup_score):
    from copy import deepcopy
    score = setup_score
    new_ts = (3, 8)
    original_bars = score.bars.copy()

    changed_score = score.change_ts(new_ts, dilate_time=False)

    for idx, bar in enumerate(changed_score.bars):
        assert bar[score.TIME_SIGNATURE_NUMERATOR_INDEX] == new_ts[0]
        assert bar[score.TIME_SIGNATURE_DENOMINATOR_INDEX] == new_ts[1]
        assert bar[score.TIME_SIGNATURE_NUMERATOR_INDEX] != original_bars[idx][score.TIME_SIGNATURE_NUMERATOR_INDEX]
        assert bar[score.TIME_SIGNATURE_DENOMINATOR_INDEX] != original_bars[idx][score.TIME_SIGNATURE_DENOMINATOR_INDEX]

def test_change_ts_with_dilate(setup_score):
    score = setup_score
    new_ts = (3, 4)
    original_times = [track["time"].copy() for track in score.tracks[score.track_keys[0]]]
    original_durations = [track["duration"].copy() for track in score.tracks[score.track_keys[0]]]

    changed_score = score.change_ts(new_ts, dilate_time=True)

    for idx, bar in enumerate(changed_score.bars):
        assert bar[score.TIME_SIGNATURE_NUMERATOR_INDEX] == new_ts[0]
        assert bar[score.TIME_SIGNATURE_DENOMINATOR_INDEX] == new_ts[1]

    dilate_factors = []
    for idx_bar, bar in enumerate(score.bars):
        original_bar_duration = score._get_bar_duration_from_bar(bar, score.tpq)
        new_bar = list(bar)
        new_bar[score.TIME_SIGNATURE_NUMERATOR_INDEX] = new_ts[0]
        new_bar[score.TIME_SIGNATURE_DENOMINATOR_INDEX] = new_ts[1]
        new_bar_duration = score._get_bar_duration_from_bar(new_bar, score.tpq)
        dilate_factors.append(new_bar_duration / original_bar_duration)

    for bar in range(len(changed_score.bars)):
        track = changed_score.tracks[changed_score.track_keys[0]][bar]
        factor = dilate_factors[bar]
        assert np.array_equal(track["time"], (original_times[bar] * factor).astype(np.int32))
        assert np.array_equal(track["duration"], (original_durations[bar] * factor).astype(np.int32))



def test_is_bar_track_empty(setup_score):
    score = setup_score
    for i in [1]:
        for key in score.tracks[score.track_keys[0]][i].keys():
            score.tracks[score.track_keys[0]][i][key] = np.array([])

    score.add_note(pitch=60, time=0, duration=3, velocity=100, track_index=0, bar_index=2)
    assert score.is_bar_track_empty(0, 1) == True # Empty bar
    assert score.is_bar_track_empty(0, 0) == True # Ghost note
    assert score.is_bar_track_empty(0, 2) == False # Normal note


def test_is_bar_empty(setup_score):
    score = setup_score.copy()
    score = score.add_instrument(instrument.PIANO)
    for i in [1]:
        for key in score.tracks[score.track_keys[0]][i].keys():
            score.tracks[score.track_keys[0]][i][key] = np.array([])

    score.add_note(pitch=60, time=0, duration=3, velocity=100, track_index=0, bar_index=2)
    assert score.is_bar_empty(1) == True  # Empty bar
    assert score.is_bar_empty(0) == True  # Ghost note
    assert score.is_bar_empty(2) == False  # Normal note




from maidi.analysis.control_features.density import DensityTagsProvider
from maidi.analysis.control_features.polyphony import MinMaxPolyphonyTagsProvider
from maidi.analysis.control_features.register import MinMaxRegisterTagsProvider
from maidi.analysis.control_features.special_notes import SpecialNotesTagsProvider





__all__ = [
    "DensityTagsProvider",
    "MinMaxPolyphonyTagsProvider",
    "MinMaxRegisterTagsProvider",
    "SpecialNotesTagsProvider",
]





class GenerationMetric:

    """
    This class is used to evaluate the quality of a generated score.
    It can be used to select score on a best fit basis or regenerate specific bars or tracks that are not satisfying.
    """

    def __init__(self, chords, tags, predicted_score):
        from maidi import TagManager, ChordManager
        self.chords = ChordManager(chords)
        self.tags = TagManager(tags)
        self.predicted_score = predicted_score
        self.new_tags = self.get_tagger().tag_score(self.predicted_score)
        self.new_chords = ChordManager(self.predicted_score.get_chords())

    @classmethod
    def get_tagger(cls):
        from maidi.analysis import ScoreTagger
        return ScoreTagger.get_base_tagger()

    def tags_recall(self):
        """
        Compute the recall of the tags in the predicted score.
        (portion of the initial tags that are in the predicted tags for each bars of each track)
        """
        initial_tags = self.tags
        recalled_tags = 0
        total_tags = 0
        for i in range(len(self.new_tags)):
            for j in range(len(self.new_tags.tags[i])):
                new_tags_locals = self.new_tags.get_tags_at_index(i, j)
                initial_tags_local = initial_tags.get_tags_at_index(i, j)
                recalled_tags += len(set(new_tags_locals).intersection(initial_tags_local))
                total_tags += len(initial_tags_local)
        return recalled_tags / total_tags


    def tag_recall_for_index(self, track_index, bar_index):
        """
        Compute the recall of the tags in the predicted score for a specific track and bar.
        (portion of the initial tags that are in the predicted tags for a specific track and bar)

        Parameters
        ----------
        track_index : int
            The index of the track to evaluate.

        bar_index : int
            The index of the bar to evaluate.
        """
        initial_tags = self.tags

        recalled_tags = 0
        total_tags = 0
        new_tags_locals = self.new_tags.get_tags_at_index(track_index, bar_index)
        initial_tags_local = initial_tags.get_tags_at_index(track_index, bar_index)
        recalled_tags += len(set(new_tags_locals).intersection(initial_tags_local))
        total_tags += len(initial_tags_local)
        if total_tags == 0:
            return 1.0

        return recalled_tags / total_tags


    def chord_recall_for_index(self, bar_index):
        """
        Compute the recall of the chords in the predicted score for a specific bar.
        (portion of the initial chords that are in the predicted chords for a specific bar)

        Parameters
        ----------
        bar_index : int
            The index of the bar to evaluate.
        """
        initial_chord = self.chords[bar_index]
        new_chord = self.new_chords[bar_index]
        if initial_chord == new_chord:
            return 1.0
        return 0.0


    def full_recall_report(self):
        """
        Compute the recall of the tags and chords in the predicted score for each track and bar.
        Return a dictionary with two keys : 'tags' and 'chords' containing the recall for each track and bar.

        Returns
        -------
        dict
            A dictionary containing the recall for each track and bar.
        """
        recall_report = {'tags': [], 'chords': []}
        for i in range(len(self.new_tags)):
            track_tags = []
            for j in range(len(self.new_tags.tags[i])):
                track_tags.append(self.tag_recall_for_index(i, j))
                if i == 0:
                    recall_report['chords'].append(self.chord_recall_for_index(j))
            recall_report['tags'].append(track_tags)
        return recall_report

    def tag_and_chord_recalls(self):
        """ Get for each tag and each chord the recall of the tags and chords in the predicted score. """
        tags_dict_recall = {}  # key: tag, value: (nb_recalled, nb_total)
        chords_dict_recall = {}  # key: chord, value: (nb_recalled, nb_total)

        # Calculate tag recalls
        for i in range(len(self.new_tags)):
            for j in range(len(self.new_tags.tags[i])):
                new_tags_locals = self.new_tags.get_tags_at_index(i, j)
                initial_tags_local = self.tags.get_tags_at_index(i, j)
                for tag in initial_tags_local:
                    if tag not in tags_dict_recall:
                        tags_dict_recall[tag] = [0, 0]
                    tags_dict_recall[tag][1] += 1
                    if tag in new_tags_locals:
                        tags_dict_recall[tag][0] += 1

        # Calculate chord recalls
        for i in range(len(self.chords)):
            initial_chord = self.chords[i]
            new_chord = self.new_chords[i]
            if initial_chord not in chords_dict_recall:
                chords_dict_recall[initial_chord] = [0, 0]
            chords_dict_recall[initial_chord][1] += 1
            if initial_chord == new_chord:
                chords_dict_recall[initial_chord][0] += 1

        return tags_dict_recall, chords_dict_recall



    def chords_recall(self):
        """
        Compute the recall of the chords in the predicted score.
        (portion of the initial chords that are in the predicted chords)
        """
        chords = self.chords
        recalled_chords = 0
        total_chords = 0
        for i in range(len(chords)):
            initial_chord = chords[i]
            new_chord = self.new_chords[i]
            if initial_chord == new_chord:
                recalled_chords += 1
            total_chords += 1

        if total_chords == 0:
            return 1.0

        return recalled_chords / total_chords

    def global_recall(self, weight_chords=0.7, weight_tags=0.3):
        """
        Compute the global recall of the predicted score.
        The global recall is a weighted average of the recall of the tags and the recall of the chords.

        Parameters
        ----------
        weight_chords : float, optional
            The weight of the recall of the chords in the global recall. The default is 0.7.

        weight_tags : float, optional
            The weight of the recall of the tags in the global recall. The default is 0.3.
        """

        recall_chords = self.chords_recall()
        recall_tags = self.tags_recall()

        return recall_chords * weight_chords + recall_tags * weight_tags



from maidi import MidiScore, INSTRUMENTS_DICT
from maidi import instrument

import pytest
import tempfile

def create_fake_midi_file(filename):
    import pretty_midi
    import numpy as np
    # Create a PrettyMIDI object
    midi_data = pretty_midi.PrettyMIDI()
    # Create an Instrument instance
    instrument1 = pretty_midi.Instrument(program=0)
    instrument2 = pretty_midi.Instrument(program=10)
    instruments = [instrument1, instrument2]

    pitches = [60, 64, 67]
    times = [(0, 1), (1, 2), (2, 3), (3, 4)]
    for instrument in instruments:
        # Iterate over note names, which will be converted to note number later
        for start, end in times:
            for pitch in pitches:
                # Create a Note instance, starting at 0s and ending at 1s
                note = pretty_midi.Note(
                    velocity=100, pitch=pitch, start=start, end=end)
                # Add it to our instrument
                instrument.notes.append(note)
        # Add the instrument to the PrettyMIDI object
        midi_data.instruments.append(instrument)
        # Write out the MIDI data
    midi_data.write(filename)


def test_cut_silence_empty():
    score = MidiScore.from_empty(
        instruments=[instrument.PIANO, instrument.VIOLIN],
        nb_bars=3,
        ts=(4, 4),
        tempo=120
    )
    assert score.cut_silence_bars_at_end().shape == (2, 3)

def test_bar_is_empty_ghost_notes():
    score = MidiScore.from_empty(
        instruments=[instrument.PIANO, instrument.VIOLIN],
        nb_bars=3,
        ts=(4, 4),
        tempo=120
    )
    score.is_bar_empty(-1, keep_ghost_note=True)


def test_empty_midi_score():
    """ """
    score = MidiScore.from_empty(
        instruments=[instrument.PIANO, instrument.VIOLIN],
        nb_bars=4,
        ts=(4, 4),
        tempo=120

    )

    assert score.shape == (2, 4)

    # Assert still same shape if writing and loading it
    new_score = MidiScore.from_base64(score.to_base64())

    assert new_score.shape == (2, 4)


def test_add_track():
    """ """
    score = MidiScore.from_empty(
        instruments=[instrument.PIANO, instrument.VIOLIN],
        nb_bars=4,
        ts=(4, 4),
        tempo=120

    )

    new_score = score.add_instrument(instrument.PIANO)

    assert new_score.shape == (3, 4)


def test_get_item_track_slice():
    """ """
    score = MidiScore.from_empty(
        instruments=[instrument.PIANO, instrument.VIOLIN],
        nb_bars=4,
        ts=(4, 4),
        tempo=120

    )
    new_score = score[1:]
    assert new_score.shape == (1, 4)

    program, is_drum = INSTRUMENTS_DICT[instrument.VIOLIN]
    assert new_score.track_keys == [(1, program, is_drum) ]
    assert len(new_score.tracks) == 1

    new_score = score[:1]
    assert new_score.shape == (1, 4)
    assert len(new_score.tracks) == 1
    program, is_drum = INSTRUMENTS_DICT[instrument.PIANO]
    assert new_score.track_keys == [(0, program, is_drum)]

    new_score = score[:]
    assert new_score.shape == (2, 4)
    assert len(new_score.tracks) == 2
    program1, is_drum1 = INSTRUMENTS_DICT[instrument.PIANO]
    program2, is_drum2 = INSTRUMENTS_DICT[instrument.VIOLIN]
    assert new_score.track_keys == [(0, program1, is_drum1), (1, program2, is_drum2)]

def test_get_item_chord_slice():
    """ """
    score = MidiScore.from_empty(
        instruments=[instrument.PIANO, instrument.VIOLIN],
        nb_bars=4,
        ts=(4, 4),
        tempo=120

    )
    new_score = score[:, 1:]
    assert new_score.shape == (2, 3)

    new_score = score[:, :1]
    assert new_score.shape == (2, 1)

    new_score = score[:, :]
    assert new_score.shape == (2, 4)


def test_set_item_single_track():
    score = MidiScore.from_empty(
        instruments=[instrument.PIANO, instrument.VIOLIN],
        nb_bars=4,
        ts=(4, 4),
        tempo=120
    )
    other_score = MidiScore.from_empty(
        instruments=[instrument.FLUTE],
        nb_bars=4,
        ts=(4, 4),
        tempo=120
    )

    # Modify the other track
    other_score.tracks[other_score.track_keys[0]][0]['time'][0] += 2

    score[0, :] = other_score
    assert score.track_keys[0] == (0, INSTRUMENTS_DICT[instrument.PIANO][0], INSTRUMENTS_DICT[instrument.PIANO][1] == 1)
    assert score.get_track_from_index(0) == other_score.get_track_from_index(0)


def test_set_item_multiple_slice():
    score = MidiScore.from_empty(
        instruments=[instrument.PIANO, instrument.VIOLIN],
        nb_bars=4,
        ts=(4, 4),
        tempo=120
    )
    other_score = MidiScore.from_empty(
        instruments=[instrument.FLUTE],
        nb_bars=3,
        ts=(4, 4),
        tempo=120
    )

    # Modify the other track
    other_score.tracks[other_score.track_keys[0]][2]['time'][0] += 2

    score[1:, 1:] = other_score
    assert score.track_keys[1] == (1, INSTRUMENTS_DICT[instrument.VIOLIN][0], INSTRUMENTS_DICT[instrument.VIOLIN][1] == 1)
    assert score[:, 1:].get_track_from_index(1) == other_score.get_track_from_index(0)

def test_set_item_invalid_type():
    score = MidiScore.from_empty(
        instruments=[instrument.PIANO, instrument.VIOLIN],
        nb_bars=4,
        ts=(4, 4),
        tempo=120
    )

    with pytest.raises(TypeError):
        score[0, 0] = "not a MidiScore"


def test_setitem_entire_track():
    score = MidiScore.from_empty(
        instruments=[instrument.PIANO, instrument.VIOLIN],
        nb_bars=4,
        ts=(4, 4),
        tempo=120
    )
    other_score = MidiScore.from_empty(
        instruments=[instrument.FLUTE],
        nb_bars=4,
        ts=(4, 4),
        tempo=120
    )

    # Modify the other track
    other_score.tracks[other_score.track_keys[0]][0]['time'][0] += 2

    score[0] = other_score
    assert score.track_keys[0] == (0, INSTRUMENTS_DICT[instrument.PIANO][0], INSTRUMENTS_DICT[instrument.PIANO][1] == 1)
    assert score.get_track_from_index(0) == other_score.get_track_from_index(0)

def test_setitem_entire_track_more_than_one_track():
    score = MidiScore.from_empty(
        instruments=[instrument.PIANO, instrument.VIOLIN],
        nb_bars=4,
        ts=(4, 4),
        tempo=120
    )
    other_score = MidiScore.from_empty(
        instruments=[instrument.PIANO, instrument.VIOLIN],
        nb_bars=4,
        ts=(4, 4),
        tempo=120
    )

    # Assert it raises an error
    with pytest.raises(ValueError) as e:
        score[0] = other_score
    assert  "The provided MidiScore must have one track" in str(e.value)


def test_setitem_entire_track_wrong_bar_number():
    score = MidiScore.from_empty(
        instruments=[instrument.PIANO, instrument.VIOLIN],
        nb_bars=4,
        ts=(4, 4),
        tempo=120
    )
    other_score = MidiScore.from_empty(
        instruments=[instrument.PIANO],
        nb_bars=3,
        ts=(4, 4),
        tempo=120
    )

    # Assert it raises an error
    with pytest.raises(ValueError) as e:
        score[0] = other_score
    assert "The provided MidiScore must have the same number of bars" in str(e.value)


def test_setitem_assign_slice():

    score = MidiScore.from_empty(
        instruments=[instrument.PIANO, instrument.VIOLIN],
        nb_bars=4,
        ts=(4, 4),
        tempo=120
    )
    other_score = MidiScore.from_empty(
        instruments=[instrument.PIANO],
        nb_bars=4,
        ts=(4, 4),
        tempo=120
    )

    # Modify the other track
    other_score.tracks[other_score.track_keys[0]][0]['time'][0] += 2

    score[1:] = other_score
    assert score.track_keys[1] == (1, INSTRUMENTS_DICT[instrument.VIOLIN][0], INSTRUMENTS_DICT[instrument.VIOLIN][1] == 1)
    assert score.get_track_from_index(1) == other_score.get_track_from_index(0)

def test_setitem_assign_slice_wrong_bar_number():
    score = MidiScore.from_empty(
        instruments=[instrument.PIANO, instrument.VIOLIN],
        nb_bars=4,
        ts=(4, 4),
        tempo=120
    )
    other_score = MidiScore.from_empty(
        instruments=[instrument.PIANO],
        nb_bars=5,
        ts=(4, 4),
        tempo=120
    )

    # Assert it raises an error
    with pytest.raises(ValueError) as e:
        score[1:] = other_score
    assert "The provided MidiScore must have the same number of bars" in str(e.value)

def test_setitem_assign_slice_wrong_track_number():
    score = MidiScore.from_empty(
        instruments=[instrument.PIANO, instrument.VIOLIN],
        nb_bars=4,
        ts=(4, 4),
        tempo=120
    )
    other_score = MidiScore.from_empty(
        instruments=[instrument.PIANO, instrument.VIOLIN],
        nb_bars=3,
        ts=(4, 4),
        tempo=120
    )

    # Assert it raises an error
    with pytest.raises(ValueError) as e:
        score[1:] = other_score
    assert "The provided MidiScore must have the same number of tracks" in str(e.value)


def test_wrong_index_type_setitem():
    score = MidiScore.from_empty(
        instruments=[instrument.PIANO, instrument.VIOLIN],
        nb_bars=4,
        ts=(4, 4),
        tempo=120
    )


    with pytest.raises(TypeError):
        score["a"] = score[1]

def test_wrong_shape_setitem_tuple():
    score = MidiScore.from_empty(
        instruments=[instrument.PIANO, instrument.VIOLIN],
        nb_bars=4,
        ts=(4, 4),
        tempo=120
    )

    with pytest.raises(ValueError) as e:
        score[0, 0] = score[1]
    assert "Shapes do not match" in str(e.value)



def test_setitem_tuple_second_dim_int():
    score = MidiScore.from_empty(
        instruments=[instrument.PIANO, instrument.VIOLIN],
        nb_bars=4,
        ts=(4, 4),
        tempo=120
    )
    score[:, 0] = score[:, 1]

    score[0, :] = score[1, :]

    score[:, :-1] = score[:, 1:]

    score[:-1, :] = score[1:, :]

def test_different_shape_of_getitem():
    score = MidiScore.from_empty(
        instruments=[instrument.PIANO, instrument.VIOLIN],
        nb_bars=4,
        ts=(4, 4),
        tempo=120
    )

    assert score[:, 1:].shape == (2, 3)

    assert score[:, :1].shape == (2, 1)

    assert score[1:2, 1:].shape == (1, 3)

    assert score[-2:-1, 1:].shape == (1, 3)

    assert score[0, :].shape == (1, 4)



def test_assign_tracks_bar_range_none():
    score = MidiScore.from_empty(
        instruments=[instrument.PIANO, instrument.VIOLIN],
        nb_bars=4,
        ts=(4, 4),
        tempo=120
    )

    other_score = MidiScore.from_empty(
        instruments=[instrument.PIANO, instrument.VIOLIN],
        nb_bars=4,
        ts=(4, 4),
        tempo=120
    )

    score.assign_tracks([0, 1], other_score, bar_range=None)


def test_get_chords_prompts():
    score = MidiScore.from_empty(
        instruments=[instrument.PIANO, instrument.VIOLIN],
        nb_bars=4,
        ts=(4, 4),
        tempo=120
    )

    score.add_note(60, 0, 10, 40, 0, 0)
    score.add_note(64, 0, 10, 40, 0, 0)
    score.add_note(67, 0, 10, 40, 0, 0)

    chord_prompt = score.get_chords()
    assert len(chord_prompt) == 4
    assert chord_prompt[0] == (0, 0, 'M', '')


def test_get_bars_int():

    score = MidiScore.from_empty(
        instruments=[instrument.PIANO, instrument.VIOLIN],
        nb_bars=4,
        ts=(4, 4),
        tempo=120
    )

    new_score = score.get_bars(1)
    assert new_score.shape == (2, 1)

def test_get_tracks_int():

    score = MidiScore.from_empty(
        instruments=[instrument.PIANO, instrument.VIOLIN],
        nb_bars=4,
        ts=(4, 4),
        tempo=120
    )

    new_score = score.get_tracks(1)
    assert new_score.shape == (1, 4)

def test_add_note():
    score = MidiScore.from_empty(
        instruments=[instrument.PIANO],
        nb_bars=4,
        ts=(4, 4),
        tempo=120,
        add_fake_notes=False
    )
    score = score.add_note(60, 3, 10, 40, 0, 1)
    track = (0, 0, 0) # Piano
    assert len(score.tracks[track][1]['time']) == 1
    assert score.tracks[track][1]['pitch'][0] == 60
    assert score.tracks[track][1]['time'][0] == 3
    assert score.tracks[track][1]['duration'][0] == 10
    assert score.tracks[track][1]['velocity'][0] == 40



def test_delete_bar():
    score = MidiScore.from_empty(
        instruments=[instrument.PIANO],
        nb_bars=4,
        ts=(4, 4),
        tempo=120,
        add_fake_notes=False
    )
    score = score.delete_bar(3)
    assert score.shape == (1, 3)


def test_get_bar_duration():
    score1 = MidiScore.from_empty(
        instruments=[instrument.PIANO],
        nb_bars=2,
        ts=(4, 4),
        tempo=120,
        add_fake_notes=False,
        tpq=24
    )
    score2 = MidiScore.from_empty(
        instruments=[instrument.PIANO],
        nb_bars=4,
        ts=(3, 4),
        tempo=120,
        add_fake_notes=False,
        tpq=24
    )
    score = score1.concatenate(score2, axis=1)

    assert score.get_bar_duration(0) == 4 * 24
    assert score.get_bar_duration(1) == 4 * 24
    assert score.get_bar_duration(2) == 3 * 24
    assert score.get_bar_duration(3) == 3 * 24


def test_write_read_empty_score():

    score = MidiScore.from_empty(
        instruments=[instrument.PIANO, instrument.VIOLIN],
        nb_bars=4,
        ts=(4, 4),
        tempo=120,
        add_fake_notes=True
    )

    with tempfile.NamedTemporaryFile(suffix='.mid', delete=True) as f:
        score.write(f.name)
        new_score = MidiScore.from_midi(f.name)
        assert new_score.nb_bars == 4



def test_prevent_silence_tags():

    score = MidiScore.from_empty(
        instruments=[instrument.PIANO, instrument.VIOLIN],
        nb_bars=4,
        ts=(4, 4),
        tempo=120,
        add_fake_notes=True
    )

    mask, tags, chords = score.get_empty_controls(prevent_silence=False)
    new_tags = MidiScore.prevent_silence_tags(tags)

    assert new_tags[0][0][0] == 'CONTROL_MIN_POLYPHONY__1'


def test_get_empty_controls():
    score = MidiScore.from_empty(
        instruments=[instrument.PIANO, instrument.VIOLIN],
        nb_bars=4,
        ts=(4, 4),
        tempo=120,
    )

    mask, tags, chords = score.get_empty_controls()
    assert mask.shape == (2, 4)
    assert len(tags) == 2
    assert len(tags[0]) == 4
    assert len(chords) == 4


def test_get_silenced_mask():
    import numpy as np
    score = MidiScore.from_empty(
        instruments=[instrument.PIANO, instrument.VIOLIN],
        nb_bars=4,
        ts=(4, 4),
        tempo=120,
    )
    score = score.add_note(60, 0, 10, 30, 1, 1)
    silence_mask = score.get_silenced_mask()
    print(silence_mask)
    assert silence_mask[1, 1] == 0

    silence_mask[1, 1] = 1
    assert np.min(silence_mask) == 1

def test_cut_silenced_bars_at_beginning_and_end():
    score = MidiScore.from_empty(
        instruments=[instrument.PIANO, instrument.VIOLIN],
        nb_bars=3,
        ts=(4, 4),
        tempo=120,
    )

    # Add a note on second bar
    score = score.add_note(60, 1, 10, 30, 0, 1)

    # Cut the first and last bar
    score = score.cut_silenced_bars_at_beginning_and_end()

    assert score.nb_bars == 1


def test_check_mask():

    score = MidiScore.from_empty(
        instruments=[instrument.PIANO, instrument.VIOLIN],
        nb_bars=3,
        ts=(4, 4),
        tempo=120,
    )

    mask = score.get_mask()


    with pytest.raises(ValueError) as e:
        score.check_mask(mask[:-1, :])
    assert "Mask number of tracks does not match score number of tracks" in str(e.value)

    with pytest.raises(ValueError) as e:
        score.check_mask(mask[:, :-1])
    assert "Mask length does not match score length" in str(e.value)

    with pytest.raises(ValueError) as e:
        score.check_mask(mask)
    assert "Mask must have at least one 1" in str(e.value)


def test_check_times():

    score = MidiScore.from_empty(

        instruments=[instrument.PIANO, instrument.VIOLIN],
        nb_bars=3,
        ts=(4, 4),
        tempo=120,
    )

    score.bars.append(score.bars[-1])

    with pytest.raises(ValueError) as e:
        score.check_times()
    assert "Track and bar length mismatch" in str(e.value)


def test_add_instrument_wrong_instrument():
    score = MidiScore.from_empty(
        instruments=[instrument.PIANO, instrument.VIOLIN],
        nb_bars=3,
        ts=(4, 4),
        tempo=120,
    )

    with pytest.raises(ValueError) as e:
        score.add_instrument("Unknown instrument")
    assert "Unknown instrument" in str(e.value)


def test_add_instrument_twice_is_ok():

    score = MidiScore.from_empty(
        instruments=[instrument.PIANO, instrument.VIOLIN],
        nb_bars=3,
        ts=(4, 4),
        tempo=120,
    )
    score = score.add_instrument(instrument.PIANO)
    assert score.nb_tracks == 3
    score = score.add_instrument(instrument.PIANO)
    assert score.nb_tracks == 4






from .parser import Parser
from .constants import *
from .score import MidiScore
from .analysis import TagsProvider, ScoreTagger, TagManager, Tags
from .chords import ChordManager
from .integrations import MusicLangAPI

__version__ = "0.15.1"


from maidi.chords.chord_inference import fast_chord_inference

def test_chord_inference():

    tracks = {(0, 0, 0): [
        {"pitch": [60, 64, 67], "time": [0, 0, 0], "duration": [1, 1, 1]},
        {"pitch": [59, 62, 67], "time": [2, 2, 2], "duration": [1, 1, 1]},
        {"pitch": [60, 64, 67], "time": [2, 2, 2], "duration": [1, 1, 1]}
    ]}

    chord_durations = [(4, 4), (4, 4), (4, 4)]
    chords = fast_chord_inference(tracks, chord_durations)
    expected_chords = [(0, 0, 'M', 0, '', 4, 4), (4, 0, 'M', -1, '6', 4, 4), (0, 0, 'M', 0, '', 4, 4)]

    assert chords == expected_chords


from maidi.chords.chord_inference_utils import SCALES


def to_pitch(type, idx, oct, chord):
    """
    Return the pitch from the note type, idx, octave and chord

    Parameters
    ----------
    type : str
        
    idx : int
        
    oct : int
        
    chord : tuple
        

    Returns
    -------

    """
    (
        roman,  #  Chord degree 0-indexed
        tonality_root,
        tonality_mode,
        chord_octave,
        real_extension,
        num,
        den,
    ) = chord
    base_scale = SCALES[tonality_mode]
    base_scale = (
        [base_scale[roman]]
        + base_scale[roman + 1 :]
        + [a + 12 for a in base_scale[:roman]]
    )
    scale = [s + tonality_root + 12 * chord_octave for s in base_scale]
    chromatic_scale = [
        s + base_scale[0] + tonality_root + 12 * chord_octave for s in range(12)
    ]
    if type == "s":
        scale = scale
    elif type == "d":
        return 60 + idx + 12 * oct
    elif type == "a":
        return 60 + idx + 12 * oct
    else:
        scale = chromatic_scale
    return 60 + scale[idx] + 12 * oct


def pitch_to_note_type(pitch, velocity, time, duration, chord, is_drum, is_absolute):
    """

    Parameters
    ----------
    pitch :
        
    velocity :
        
    time :
        
    duration :
        
    chord :
        
    is_drum :
        
    is_absolute :
        

    Returns
    -------

    """
    type, idx, oct = parse_pitch(pitch, chord, is_drum, is_absolute)
    amp = get_amp_figure(velocity)
    time = int(time)  # Previously : - chord_start
    duration = int(duration)

    return type, idx, oct, amp, time, duration


def get_amp_figure(velocity):
    """

    Parameters
    ----------
    velocity :
        

    Returns
    -------

    """
    n = velocity / 120
    if n is None or n <= 0:
        return "n"
    elif n <= 0.16:
        return "ppp"
    elif n <= 0.26:
        return "pp"
    elif n <= 0.36:
        return "p"
    elif n <= 0.5:
        return "mp"
    elif n <= 0.65:
        return "mf"
    elif n <= 0.8:
        return "f"
    elif n <= 0.9:
        return "ff"
    else:
        return "fff"


def amp_figure_to_velocity(amp_figure):
    """

    Parameters
    ----------
    amp_figure :
        

    Returns
    -------

    """
    if amp_figure == "n":
        return 0
    elif amp_figure == "ppp":
        return int(120 * 0.16)
    elif amp_figure == "pp":
        return int(120 * 0.26)
    elif amp_figure == "p":
        return int(120 * 0.36)
    elif amp_figure == "mp":
        return int(120 * 0.5)
    elif amp_figure == "mf":
        return int(120 * 0.65)
    elif amp_figure == "f":
        return int(120 * 0.8)
    elif amp_figure == "ff":
        return int(120 * 0.9)
    else:
        return int(120 * 0.95)


def parse_pitch(pitch, chord, is_drum, is_absolute):
    """Parse an integer pitch (0=C5)

    Parameters
    ----------
    pitch :
        
    chord :
        
    is_drum :
        
    is_absolute :
        

    Returns
    -------

    
    """
    (
        roman,
        tonality_root,
        tonality_mode,
        chord_octave,
        real_extension,
        num,
        den,
    ) = chord
    pitch = pitch - 60  # C5
    if is_drum:
        return "d", pitch % 12, pitch // 12
    if is_absolute:
        return "a", pitch % 12, pitch // 12

    base_scale = SCALES[tonality_mode]
    base_scale = (
        [base_scale[roman]]
        + base_scale[roman + 1 :]
        + [a + 12 for a in base_scale[:roman]]
    )
    scale = [s + tonality_root + 12 * chord_octave for s in base_scale]
    scale_class = [s % 12 for s in scale]
    chromatic_scale = [
        s + base_scale[0] + tonality_root + 12 * chord_octave for s in range(12)
    ]
    scale_class_set = set(scale_class)
    if pitch % 12 in scale_class_set:
        scale = scale
        type = "s"
    else:
        scale = chromatic_scale
        type = "h"

    scale_mod = [s % 12 for s in scale]
    idx = scale_mod.index(pitch % 12)
    oct = (pitch - scale[0]) // 12

    return type, idx, oct



import pytest
import numpy as np
from unittest.mock import patch, MagicMock
from maidi import MidiScore
from maidi.integrations.api import MusicLangAPI
from maidi import instrument

API_URL = "https://api.mockmusiclang.io"
MUSICLANG_API_KEY = "testapikey"


@pytest.fixture
def setup_api():
    return MusicLangAPI(api_url=API_URL, api_key=MUSICLANG_API_KEY, verbose=True)


@pytest.fixture
def setup_score():
    instruments = [instrument.PIANO]
    score = MidiScore.from_empty(instruments=instruments, nb_bars=8, ts=(4, 4), tempo=120)
    return score


def test_determine_bars_step(setup_api):
    api = setup_api

    step = api._determine_bars_step(nb_added_bars_step=4, remaining_bars=10)
    assert step == 4

    step = api._determine_bars_step(nb_added_bars_step=None, remaining_bars=10)
    assert step == min(api.MAX_CONTEXT, 10)



@patch.object(MusicLangAPI, 'predict')
def test_generate_from_scratch(mock_predict, setup_api, setup_score):
    api = setup_api
    score = setup_score

    def predict_side_effect(predicted_score, mask, model, tags, chords, temperature, polling_interval, timeout,
                            **prediction_kwargs):
        return predicted_score  # Return the input segment as the result

    mock_predict.side_effect = predict_side_effect

    generated_score = api.generate_from_scratch(instruments=["piano"], nb_bars=8, ts=(4, 4), tempo=120, temperature=0.95,
                                               model="control_masking_large", timeout=120, polling_interval=1)

    assert mock_predict.called
    assert generated_score is not None
    assert generated_score.nb_bars == 8



if __name__ == "__main__":
    pytest.main()



from maidi.analysis.parse_notes import parse_pitch, to_pitch, get_amp_figure, amp_figure_to_velocity, pitch_to_note_type

###################

## Write the test functions below

def test_parse_pitch():
    assert parse_pitch(60, (0, 2, 'm', 0, '', 4, 4), False, False) == ('h', 10, -1)
    assert parse_pitch(60, (0, 2, 'm', 0, '', 4, 4), False, True) == ('a', 0, 0)
    assert parse_pitch(61, (0, 2, 'm', 0, '', 4, 4), True, False) == ('d', 1, 0)
    assert parse_pitch(50, (0, 2, 'm', 0, '', 4, 4), False, False) == ('s', 0, -1)

def test_to_pitch():

    assert to_pitch('s', 1, -1, (0, 0, 'm', 0, '', 4, 4)) == 50
    assert to_pitch('h', 10, -1, (0, 0, 'm', 0, '', 4, 4)) == 58
    assert to_pitch('a', 0, 0, (0, 1, 'm', 0, '', 4, 4)) == 60
    assert to_pitch('d', 1, 0, (0, 1, 'm', 0, '', 4, 4)) == 61


def test_get_amp_figure():

    assert get_amp_figure(0) == "n"
    assert get_amp_figure(127) == 'fff'

    velocity = amp_figure_to_velocity("f")
    assert amp_figure_to_velocity(get_amp_figure(velocity))  == velocity


def test_pitch_to_note_type():

    pitch = 60
    velocity = 127
    time = 0
    duration = 1
    chord = (0, 2, 'm', 0, '', 4, 4)
    is_drum = False
    is_absolute = False

    type, idx, oct, amp, time, duration = pitch_to_note_type(pitch, velocity, time, duration, chord, is_drum, is_absolute)

    assert (type, idx, oct, amp, time, duration) == ('h', 10, -1, 'fff', 0, 1)



import re
import itertools

VALID_EXTENSIONS_ = {'6', '64', '7', '65', '43', '2', ''}
VALID_SUS = {'', '(sus2)', '(sus4)'}
VALID_EXTENSIONS = set([a + b for a, b in itertools.product(VALID_EXTENSIONS_, VALID_SUS)])
VALID_ADDED_NOTES = {'[add9]', '[addb9]', '[add11]', '[add#11]', '[add13]', '[addb13]', '[add#13]', '[add#9]', '[9]', '[b9]', '[11]',
                     '[#11]', '[13]', '[b13]', '[#13]', '[#9]'}


CHORD_REGEX = re.compile(r"^(Ger|Cad|Fr|It|N|[#b]*[ivIV]+[o%ø+]?)(6|64|7|65|43|2)?(\(sus2\)|\(sus4\))?(\[add9\]|\[addb9\]|\[add11\]|\[add#11\]|\[add13\]|\[addb13\]|\[add#10\]|\[add#13\]|\[add#9\]|\[9\]|\[b9\]|\[11\]|\[#11\]|\[13\]|\[b13\]|\[#13\|\[#10\]])?(/[#b]*[ivIV]+)?$")
CHORD_PROGRESSION_REGEX = re.compile(r"(x\s*)?([A-Ga-g][#b]?:\s*)?((x|Ger|Cad|Fr|It|N|[#b]*[ivIV]+[o%ø+]?)(6|64|7|65|43|2)?(\(sus2\)|\(sus4\))?(\[add9\]|\[addb9\]|\[add11\]|\[add#11\]|\[add13\]|\[addb13\]|\[add#10\]|\[add#13\]|\[add#9\]|\[9\]|\[b9\]|\[11\]|\[#11\]|\[13\]|\[b13\]|\[#13\|\[#10\])?(/[#b]*[ivIV]+)?\s*)+")

def split_by_tonality(input_string):
    """
    Splits the input string by tonalities and associates each chord with its primary tonality.

    Args:
        input_string (str): The input string containing chords with tonalities.

    Returns:
        List[Tuple[str, str]]: A list of tuples where each tuple contains a chord and its associated primary tonality.
    """
    input_string = " ".join(input_string.split())
    # Regular expression to match tonalities (e.g., C:, c:, d#:)
    tonality_pattern = re.compile(r'([A-Ga-g][#b]?:)')

    # Find all tonalities in the input string
    tonalities = tonality_pattern.findall(input_string)
    if not tonalities:
        raise ValueError("No tonalities found in the input string.")

    # Split the input string by tonalities
    segments = tonality_pattern.split(input_string)

    # Remove empty strings from segments
    segments = [seg for seg in segments if seg]

    if len(segments) % 2 != 0:
        raise ValueError("Malformed input string. Tonalities and chords are not properly paired.")

    # Pair each chord segment with its associated tonality
    result = []
    for i in range(0, len(segments), 2):
        tonality = segments[i]
        chords = segments[i + 1].strip().split()
        for chord in chords:
            result.append((chord, tonality))

    return result


TONALITIES_DICT = {
    "c": 0, "d": 2, "e": 4, "f": 5, "g": 7, "a": 9, "b": 11
}


def tonality_name_to_degree(main_tonality_degree):
    main_tonality_degree = main_tonality_degree.lower()
    main_tonality_note_degree = TONALITIES_DICT[main_tonality_degree[0]]

    # Count the number of sharps in tonality
    nb_sharps = re.findall("#", main_tonality_degree[1:])
    nb_sharps = len(nb_sharps)

    # Count the number of flats in tonality
    nb_flats = re.findall("b", main_tonality_degree[1:])
    nb_flats = len(nb_flats)

    # Determine the tonality degree modified with flats and sharps
    main_tonality_note_degree = (main_tonality_note_degree + nb_sharps - nb_flats) % 12

    return main_tonality_note_degree


def parse_special_cases(chord_result):
    chord_degree = chord_result['chord_degree']
    chord_extension = chord_result['chord_extension']

    if chord_degree == 'N' and chord_extension == '':
        chord_result['chord_extension'] = '6'
        chord_result['final_extension'] = '6'

    if chord_degree == 'Cad' and chord_extension == '':
        chord_result['chord_extension'] = '64'
        chord_result['final_extension'] = '64'

    return chord_result


def parse_chord_dict(chord, tonality):
    """
    Parses a chord and extracts the desired properties.

    Args:
        chord (str): The chord to be parsed.
        tonality (str): The primary tonality associated with the chord.

    Returns:
        Dict[str, str]: A dictionary containing the parsed properties of the chord.
    """
    # Regular expressions for different parts of the chord
    degree_pattern = re.compile(r'^(Ger|Cad|Fr|It|N|(#|b)*[ivIV]+)')
    secondary_tonality_pattern = re.compile(r'/((#|b)*[ivIV]+)')

    # Extract chord degree
    degree_match = degree_pattern.match(chord)
    if not degree_match:
        raise ValueError(f"Invalid chord degree in chord: {chord}")
    chord_degree = degree_match.group(0)

    # Extract chord extension
    extension_match = chord.split('/')[0]
    chord_extension = extension_match[len(chord_degree):]

    # Extract secondary tonality
    secondary_tonality_match = secondary_tonality_pattern.search(chord)
    secondary_tonality = secondary_tonality_match.group(1) if secondary_tonality_match else None

    # Determine main tonality degree and mode
    main_tonality_degree = tonality[:-1]  # Remove the colon
    main_tonality_degree = tonality_name_to_degree(main_tonality_degree)
    main_tonality_mode = "m" if tonality[0].islower() else "M"

    result = {
        "chord_degree": chord_degree,
        "chord_extension": chord_extension,
        "secondary_tonality": secondary_tonality,
        "main_tonality_degree": main_tonality_degree,
        "main_tonality_mode": main_tonality_mode
    }

    from maidi.chords.chord_analyzer.chord_constants import DICT_TONALITY, DICT_RELATIVE_CHANGE

    # Determine the actual tonality degree if secondary tonality
    if secondary_tonality:
        tonality_degree, tonality_mode = DICT_TONALITY[main_tonality_mode][secondary_tonality]
        result["real_tonality_degree"] = (tonality_degree + main_tonality_degree) % 12
        result["real_tonality_mode"] = tonality_mode
    else:
        result["real_tonality_degree"] = main_tonality_degree
        result["real_tonality_mode"] = main_tonality_mode

    # Get the extension only if "o" or "%" or "ø" or "+" is present
    chord_extension_found = re.search(r'[o%ø+]', chord_extension)
    chord_extension_found = chord_extension_found.group(0) if chord_extension_found else ""

    # Remove chord extension found from extension
    result['final_extension'] = result['chord_extension'].replace(chord_extension_found, "")

    total_degree = chord_degree + chord_extension_found
    total_degree = total_degree.replace('%', 'ø')

    # Determine the relative change (0, 'M', 0), Degree, mode, offset tonality
    real_tonality_mode = result['real_tonality_mode']
    degree, final_mode, offset_tonality = DICT_RELATIVE_CHANGE[real_tonality_mode][total_degree]
    result['final_tonality_degree'] = (result['real_tonality_degree'] + offset_tonality) % 12
    result['final_tonality_mode'] = final_mode
    result['final_chord_degree'] = degree

    result = parse_special_cases(result)

    return result


def parse_chord(chord, tonality):
    """
    Parses a chord and extracts the desired properties.

    Args:
        chord (str): The chord to be parsed.
        tonality (str): The primary tonality associated with the chord.

    Returns:
        Tuple: A tuple containing the parsed properties of the chord.
    """
    # Check valid chord
    if chord == 'x':
        return None
    if not CHORD_REGEX.match(chord):
        raise ValueError(f"Invalid chord: {chord}")

    try:
        result = parse_chord_dict(chord, tonality)
        degree = result['final_chord_degree']
        tonality = result['final_tonality_degree']
        mode = result['final_tonality_mode']
        extension = result['final_extension']

        # Split the extension and added notes
        final_roman_extension, final_added_notes = split_extension_and_added_notes(extension)

        return degree, tonality, mode, final_roman_extension, final_added_notes
    except ValueError as e:
        raise ValueError(f"Error parsing chord: {e}")



def split_extension_and_added_notes(extension):
    """
    Splits the extension into roman extension and added notes.

    Args:
        extension (str): The chord extension.

    Returns:
        Tuple[str, List[str]]: A tuple containing the roman extension and a list of added notes.
    """

    remove_roman_extension = False

    # Split the extension by parentheses
    parts = re.split(r'(\[.*?\])', extension)

    final_roman_extension = ""
    final_added_notes = []

    for part in parts:
        # This is the main extension part
        if part in VALID_EXTENSIONS:
            final_roman_extension += part
        elif part == "":
            # Handle the case where the roman numeral extension is an empty string
            continue
        else:
            # Check if the part is an added note without parentheses
            if part in VALID_ADDED_NOTES:
                final_added_notes.append(part)
            else:
                raise ValueError(f"Invalid roman extension: {part}")

    if remove_roman_extension:
        final_roman_extension = ""

    return final_roman_extension, final_added_notes
def parse_roman_numeral_notation(input_string):
    """
    Parses a Roman numeral notated string and extracts the desired properties for each chord.

    Args:
        input_string (str): The input string containing chords with tonalities.

    Returns:
        List[Dict[str, str]]: A list of dictionaries where each dictionary contains the parsed properties of a chord.
    """
    # Check chord progression regex
    input_string = " ".join(input_string.split())
    if not CHORD_PROGRESSION_REGEX.match(input_string):
        raise ValueError(f"Invalid chord progression: {input_string}")

    try:
        # Split the input string by tonalities
        chords_with_tonalities = split_by_tonality(input_string)

        # Parse each chord and extract properties
        parsed_chords = []
        for chord, tonality in chords_with_tonalities:
            parsed_chord = parse_chord(chord, tonality)
            parsed_chords.append(parsed_chord)

        return parsed_chords
    except ValueError as e:
        raise ValueError(f"Error parsing input string: {e}")


from .chord_analyzer import parse_roman_numeral_notation
from maidi.analysis import Tags
from copy import deepcopy

ADDED_NOTE_TO_TAG = {

    '[addb9]': Tags.special_notes.diminished_ninth,
    '[add9]': Tags.special_notes.ninth,
    '[add#9]': Tags.special_notes.minor_third,
    '[add#10]': Tags.special_notes.major_third,
    '[add11]': Tags.special_notes.eleventh,
    '[addb5]': Tags.special_notes.diminished_fifth,
    '[add#11]': Tags.special_notes.diminished_fifth,
    '[add6]': Tags.special_notes.sixth,
    '[addb6]': Tags.special_notes.minor_sixth,
    '[add#6]': Tags.special_notes.major_sixth,
    '[addb13]': Tags.special_notes.minor_sixth,
    '[add13]': Tags.special_notes.sixth,
    '[add#13]': Tags.special_notes.major_sixth,
    '[addb7]': Tags.special_notes.minor_seventh,
    '[add7]': Tags.special_notes.major_seventh,
}

class ChordManager:

    tags = Tags

    def __init__(self, chords):

        if isinstance(chords, ChordManager):
            chords = chords.to_chords()

        self.chords = []
        for chord in chords:
            if chord is None:
                chord = self.get_none_chord()
            elif len(chord) == 4:
                chord = (*chord, [])
            self.chords.append(chord)

    def get_chord(self, bar):

        return self.chords[bar]

    def __getitem__(self, item):
        if isinstance(item, slice):
            return ChordManager(self.chords[item])
        else:
            chord = self.chords[item]
            return self.to_chord(chord)

    def __iter__(self):
        return iter(self.chords)

    @classmethod
    def to_chord(self, chord):
        return tuple(list(chord)[:4]) if chord[0] is not None else None

    @classmethod
    def get_none_chord(cls):
        return (None, None, None, [])

    def __setitem__(self, key, value):

            if isinstance(key, slice):
                if isinstance(value, (tuple, list)) and not isinstance(value[0], (tuple, list)):
                    value = [value for _ in range(*key.indices(len(self.chords)))]
                elif value is None:
                    value = [self.get_none_chord() for _ in self.chords[key]]
                elif isinstance(value, list):
                    indices = range(*key.indices(len(self.chords)))
                    expected_length = len(indices)
                    if len(value) != expected_length:
                        raise ValueError(f'Wrong number of chords assigned, expected {len(key.indices(len(self.chords)))}, got {len(value)}')
                    value = [self.get_none_chord() if chord is None else chord for chord in value]
                    for indice, val in zip(indices, value):
                        self[indice] = val

            elif isinstance(key, int):
                if value is None:
                    value = self.get_none_chord()
                elif isinstance(value, str):
                    value = parse_roman_numeral_notation(value)[0]

            print(value)
            self.chords[key] = value

    def __len__(self):
            return len(self.chords)

    def __repr__(self):
        return f'ChordManager({self.chords})'

    def to_chords(self):
        return deepcopy([self.to_chord(chord) for chord in self.chords])

    @classmethod
    def parse_added_notes_as_tags(cls, added_notes):
        tags = []
        for note in added_notes:
            tag = ADDED_NOTE_TO_TAG.get(note)
            if tag is not None:
                tags.append(tag)
        return tags

    @classmethod
    def from_roman_string(cls, roman_string):

        chords_candidate = parse_roman_numeral_notation(roman_string)

        chords = []
        for chord in chords_candidate:
            if chord is None:
                chords.append(None)
            else:
                degree, tonality, mode, extension, added_notes = chord
                chords.append((degree, tonality, mode, extension, cls.parse_added_notes_as_tags(added_notes)))

        return cls(chords)


import pytest
from unittest.mock import MagicMock, patch

# Mocking the ChordManager and TagManager classes
class MockChordManager:
    def __init__(self, chords):
        self.chords = chords

    def __len__(self):
        return len(self.chords)

    def __getitem__(self, item):
        return self.chords[item]

class MockTagManager:
    def __init__(self, tags):
        self.tags = tags

    def get_tags_at_index(self, track_index, bar_index):
        return self.tags[track_index][bar_index]

    def __len__(self):
        return len(self.tags)

    def __getitem__(self, item):
        return self.tags[item]


# Mocking the ScoreTagger class and its method
class MockScoreTagger:
    @staticmethod
    def get_base_tagger():
        return MockScoreTagger()

    def tag_score(self, predicted_score):
        return predicted_score

# Patching the imports in the GenerationMetric class
@patch('maidi.TagManager', MockTagManager)
@patch('maidi.ChordManager', MockChordManager)
@patch('maidi.analysis.ScoreTagger', MockScoreTagger)
def test_tags_recall():
    from maidi.analysis.metrics import GenerationMetric  # Replace 'your_module' with the actual module name

    initial_tags = [
        [['tag1', 'tag2'], ['tag3']],
        [['tag4'], ['tag5', 'tag6']]
    ]
    predicted_tags = [
        [['tag1'], ['tag3', 'tag7']],
        [['tag4'], ['tag5']]
    ]
    predicted_score = MagicMock()
    predicted_score.get_tags_at_index = lambda i, j: predicted_tags[i][j]

    metric = GenerationMetric(initial_tags, initial_tags, predicted_score)
    metric.new_tags = MockTagManager(predicted_tags)


    recall = metric.tags_recall()
    assert recall == 4 / 6  # 4 recalled tags out of 6 total tags

@patch('maidi.TagManager', MockTagManager)
@patch('maidi.ChordManager', MockChordManager)
@patch('maidi.analysis.ScoreTagger', MockScoreTagger)
def test_tag_recall_for_index():
    from maidi.analysis.metrics import GenerationMetric

    initial_tags = [
        [['tag1', 'tag2'], ['tag3']],
        [['tag4'], ['tag5', 'tag6']]
    ]
    predicted_tags = [
        [['tag1'], ['tag3', 'tag7']],
        [['tag4'], ['tag5']]
    ]
    predicted_score = MagicMock()
    predicted_score.get_tags_at_index = lambda i, j: predicted_tags[i][j]

    metric = GenerationMetric(initial_tags, initial_tags, predicted_score)
    metric.new_tags = MockTagManager(predicted_tags)

    recall = metric.tag_recall_for_index(0, 0)
    assert recall == 1 / 2  # 1 recalled tag out of 2 total tags

@patch('maidi.TagManager', MockTagManager)
@patch('maidi.ChordManager', MockChordManager)
@patch('maidi.analysis.ScoreTagger', MockScoreTagger)
def test_chords_recall():
    from maidi.analysis.metrics import GenerationMetric

    initial_chords = ['C', 'G', 'Am', 'F']
    predicted_chords = ['C', 'G', 'Em', 'F']
    predicted_score = MagicMock()
    predicted_score.get_chords = lambda: predicted_chords

    metric = GenerationMetric(initial_chords, initial_chords, predicted_score)
    metric.new_chords = predicted_chords
    recall = metric.chords_recall()
    assert recall == 3 / 4  # 3 recalled chords out of 4 total chords

@patch('maidi.TagManager', MockTagManager)
@patch('maidi.ChordManager', MockChordManager)
@patch('maidi.analysis.ScoreTagger', MockScoreTagger)
def test_global_recall():
    from maidi.analysis.metrics import GenerationMetric

    initial_tags = [
        [['tag1', 'tag2'], ['tag3']],
        [['tag4'], ['tag5', 'tag6']]
    ]
    predicted_tags = [
        [['tag1'], ['tag3', 'tag7']],
        [['tag4'], ['tag5']]
    ]
    initial_chords = ['C', 'G', 'Am', 'F']
    predicted_chords = ['C', 'G', 'Em', 'F']
    predicted_score = MagicMock()
    predicted_score.get_tags_at_index = lambda i, j: predicted_tags[i][j]
    predicted_score.get_chords = lambda: predicted_chords

    metric = GenerationMetric(initial_chords, initial_tags, predicted_score)
    metric.new_tags = MockTagManager(predicted_tags)
    metric.new_chords = predicted_chords
    global_recall = metric.global_recall()
    expected_recall = (3 / 4) * 0.7 + (4 / 6) * 0.3  # Weighted average of chords and tags recall
    assert global_recall == expected_recall


@patch('maidi.TagManager', MockTagManager)
@patch('maidi.ChordManager', MockChordManager)
@patch('maidi.analysis.ScoreTagger', MockScoreTagger)
def test_full_recall_report():
    from maidi.analysis.metrics import GenerationMetric

    initial_tags = [
        [['tag1', 'tag2'], ['tag3']],
        [['tag4'], ['tag5', 'tag6']]
    ]
    predicted_tags = [
        [['tag1'], ['tag3', 'tag7']],
        [['tag4'], ['tag5']]
    ]
    initial_chords = ['C', 'A']
    predicted_chords = ['C', 'G']
    predicted_score = MagicMock()
    predicted_score.get_tags_at_index = lambda i, j: predicted_tags[i][j]
    predicted_score.get_chords = lambda: predicted_chords

    metric = GenerationMetric(initial_chords, initial_tags, predicted_score)
    metric.new_tags = MockTagManager(predicted_tags)
    metric.new_chords = predicted_chords
    recall_report = metric.full_recall_report()

    expected_tags_recall = [
        [1 / 2,  # Track 0, Bar 0
        1 / 1],  # Track 0, Bar 1
        [1 / 1,  # Track 1, Bar 0
        1 / 2]  # Track 1, Bar 1
    ]
    expected_chords_recall = [
        1 / 1,  # Bar 0
        0 / 1,  # Bar 1
    ]

    assert recall_report['tags'] == expected_tags_recall
    assert recall_report['chords'] == expected_chords_recall

@patch('maidi.TagManager', MockTagManager)
@patch('maidi.ChordManager', MockChordManager)
@patch('maidi.analysis.ScoreTagger', MockScoreTagger)
def test_chords_recall():
    from maidi.analysis.metrics import GenerationMetric

    initial_chords = ['C', 'G', 'Am', 'F']
    predicted_chords = ['C', 'G', 'Em', 'F']
    predicted_score = MagicMock()
    predicted_score.get_chords = lambda: predicted_chords

    metric = GenerationMetric(initial_chords, initial_chords, predicted_score)

    metric.new_chords = predicted_chords


    recall = metric.chords_recall()
    assert recall == 3 / 4  # 3 recalled chords out of 4 total chords


import pytest
from unittest.mock import MagicMock, patch

# Mocking the ChordManager and TagManager classes
class MockChordManager:
    def __init__(self, chords):
        self.chords = chords

    def __len__(self):
        return len(self.chords)

    def __getitem__(self, item):
        return self.chords[item]


# Patching the imports in the GenerationMetric class
@patch('maidi.TagManager', MockTagManager)
@patch('maidi.ChordManager', MockChordManager)
@patch('maidi.analysis.ScoreTagger', MockScoreTagger)
def test_tag_and_chord_recalls():
    from maidi.analysis.metrics import GenerationMetric  # Replace 'your_module' with the actual module name

    initial_tags = [
        [['tag1', 'tag2'], ['tag3', 'tag1']],
        [['tag4'], ['tag5', 'tag6']]
    ]
    predicted_tags = [
        [['tag1'], ['tag3', 'tag7']],
        [['tag4'], ['tag5']]
    ]
    initial_chords = [('C', 'G', 'major', '7'), ('G', 'D', 'minor', '7')]
    predicted_chords = [('C', 'G', 'major', '7'), ('G', 'D', 'major', '7')]

    predicted_score = MagicMock()
    predicted_score.get_tags_at_index = lambda i, j: predicted_tags[i][j]
    predicted_score.get_chords = lambda: predicted_chords

    metric = GenerationMetric(initial_chords, initial_tags, predicted_score)
    metric.new_tags = MockTagManager(predicted_tags)
    metric.new_chords = MockChordManager(predicted_chords)

    tags_dict_recall, chords_dict_recall = metric.tag_and_chord_recalls()

    expected_tags_dict_recall = {
        'tag1': [1, 2],
        'tag2': [0, 1],
        'tag3': [1, 1],
        'tag4': [1, 1],
        'tag5': [1, 1],
        'tag6': [0, 1]
    }
    expected_chords_dict_recall = {
        ('C', 'G', 'major', '7'): [1, 1],
        ('G', 'D', 'minor', '7'): [0, 1]
    }

    assert tags_dict_recall == expected_tags_dict_recall
    assert chords_dict_recall == expected_chords_dict_recall


if __name__ == "__main__":
    pytest.main()


from maidi.utils.bar_helpers import bar_to_bar_duration_in_ticks, bar_to_bar_duration_in_quarters


def test_bar_to_bar_duration_in_ticks():
    bar = [0, 0, 'm', -1, '', 3, 4, 0] # 3/4
    assert bar_to_bar_duration_in_ticks(bar, tpq=24) == 72

def test_bar_to_bar_duration_in_quarters():
    bar = [0, 0, 'm', -1, '', 4, 4, 0]
    assert bar_to_bar_duration_in_quarters(bar) == 4

    bar = [0, 0, 'm', -1, '', 3, 8, 0]
    assert bar_to_bar_duration_in_quarters(bar) == 1.5




from maidi.analysis import TagsProvider


MAX_MIN_POLYPHONY = 6
MAX_MAX_POLYPHONY = 6

def get_tags_names():
    """

    Returns
    -------

    """
    return [
        "CONTROL_MIN_POLYPHONY__" + str(i) for i in range(MAX_MIN_POLYPHONY+1)
    ] + [
        "CONTROL_MAX_POLYPHONY__" + str(i) for i in range(MAX_MAX_POLYPHONY+1)
    ]


class MinMaxPolyphonyTagsProvider(TagsProvider):
    ALL_TAGS = get_tags_names()

    def get_tags(self, track_bar, chord, score):
        notes = self.get_start_end_notes(track_bar, chord, score)
        return get_min_max_polyphony_tags(notes)

def get_max_polyphony(notes):
    """

    Parameters
    ----------
    notes :
        

    Returns
    -------

    """
    if not notes:
        return 0

    events = []
    for start, end in notes:
        events.append((start, "start"))
        events.append((end, "end"))

    # Sort events, ensuring end events come before start events if they are at the same time
    events.sort(key=lambda x: (x[0], x[1] == "start"))

    max_count = 0
    current_count = 0

    for time, event_type in events:
        if event_type == "start":
            current_count += 1
            max_count = max(max_count, current_count)
        elif event_type == "end":
            current_count -= 1

    return min(max_count, MAX_MAX_POLYPHONY)


def get_min_polyphony(notes):
    """

    Parameters
    ----------
    notes :
        

    Returns
    -------

    """
    if not notes:
        return 0

    events = []
    for start, end in notes:
        events.append((start, "start"))
        events.append((end, "end"))

    # Sort events with end events before start if at the same time to ensure a note ending and starting at the same time is not counted twice
    events.sort(key=lambda x: (x[0], x[1] == "start"))

    # (1, start), (1, start), (1, start), (1, end), (1, end), (1, end)
    current_count = 0
    min_count = float("inf")
    prev_time = 0
    # Process each event
    for time, event_type in events:
        if event_type == "start":
            current_count += 1
        else:  # event_type == 'end'
            # Check for non-zero counts to track the minimum before decrementing
            if current_count > 0 and prev_time != time:
                min_count = min(min_count, current_count)
                prev_time = time
            current_count -= 1

    # If min_count was never updated, return 0, otherwise, return min_count
    return min(min_count, MAX_MIN_POLYPHONY) if min_count != float("inf") else 0




def get_min_polyphony_tags(notes):
    """

    Parameters
    ----------
    notes :
    """

    return ["CONTROL_MIN_POLYPHONY__" + str(get_min_polyphony(notes))]

def get_max_polyphony_tags(notes):
    """

    Parameters
    ----------
    notes :


    Returns
    -------

    """
    return ["CONTROL_MAX_POLYPHONY__" + str(get_max_polyphony(notes))]

def get_min_max_polyphony_tags(notes):
    """

    Parameters
    ----------
    notes :


    Returns
    -------

    """
    return get_min_polyphony_tags(notes) + get_max_polyphony_tags(notes)


MODEL_CONTROL_MASKING_LARGE = "control_masking_large"
MODEL_MASKING_LARGE = "masking_large"
MODEL_MASKING_SMALL = "masking_small"
MODEL_CONTROL_MASKING_OPTIMIZED = "control_masking_optimized"

MODELS = [MODEL_CONTROL_MASKING_LARGE, MODEL_MASKING_LARGE, MODEL_MASKING_SMALL, MODEL_CONTROL_MASKING_OPTIMIZED]
ONLY_SYNC_MODELS = []



import symusic
import numpy as np
from fractions import Fraction as frac
from symusic import Track, TimeSignature, Note, Score, Tempo
import time
from maidi.chords.chord_inference import fast_chord_inference


class Parser:
    """
    Low level parser class to read and write midi files
    You probably don't need to use this class directly, use MidiScore instead

    Usage:
    ------
    ```python

    from maidi.parser import Parser
    parser = Parser()
    chords, tracks, track_keys, tempo = parser.parse("input.mid")
    score = parser.write(chords, tracks, track_keys, tempo)
    score.dump_midi("output.mid")

    ```
    """

    def __init__(self, verbose=False, tpq=24, min_dur=1):
        """
        Initialize the parser
        Parameters
        ----------
        verbose : bool
            Whether to print debug information
        tpq : int (Default value = 24)
            Ticks per quarter note for the midi file, it will be resampled to this value
        min_dur : int (Default value = 1)
            Minimum duration for a note in ticks
        """
        self.tpq = tpq
        self.min_dur = min_dur
        self.verbose = verbose

    def parse(self, midi_file, chord_range=None, force_ts=None):
        """

        Parameters
        ----------
        midi_file : str
            Path to the midi file to parse

        chord_range : tuple or None (Default value = None)
            Tuple with the start and end chord index to parse. If None, parse the whole file

        force_ts : tuple or None (Default value = None)
            Tuple with the time signature to force for the whole file. If None, use the time signature of the midi file

        Returns
        -------
        chords : list
            List of chords with the time signature of each bar
        tracks : dict
            Dict of arrays with the notes for each track
        track_keys : list
            List of tuples with the track keys (program, is_drum, voice)


        """
        start = time.time()
        score_quantized = self._preload_score(midi_file)
        tempo = (
            score_quantized.tempos[0].qpm if len(score_quantized.tempos) > 0 else 120
        )

        self.pprint("Resampling time: ", time.time() - start)
        start = time.time()
        bars, chord_durations = self._get_bars(score_quantized, force_ts)
        self.pprint("Getting bars time: ", time.time() - start)
        tracks = {}
        track_keys = []
        tracks_global = {}
        start = time.time()

        if chord_range is None:
            chord_range = (0, len(bars))

        for idx, track in enumerate(score_quantized.tracks):
            is_drum = track.is_drum
            program = track.program if not is_drum else 0
            score_soa = track.notes.numpy()  # Convert to a dict of arrays
            track_keys.append((idx, program, is_drum))
            # Get track between chord range
            if chord_range is not None:
                start_time = bars[chord_range[0]][0]
                last_chord_index = min(chord_range[1], len(bars)) - 1
                end_time = bars[last_chord_index][1]
                mask = (score_soa["time"] >= start_time) & (
                        score_soa["time"] < end_time
                )
                for k, v in score_soa.items():
                    score_soa[k] = v[mask]

            score_soa["bar"] = self._get_bar_nb_array_optimized(score_soa, bars)
            score_soa["start_bar_tick"] = np.array(
                [bars[bar_nb][0] for bar_nb in score_soa["bar"]]
            )
            score_soa["time"] -= score_soa["start_bar_tick"]  # Make time relative to the start of the bar
            del score_soa["start_bar_tick"]
            # Group SOA by bar
            grouped_score = self._groupby_bar(score_soa, bars)
            # If chord range is specified, keep only the specified chords
            if chord_range is not None:
                grouped_score = grouped_score[chord_range[0]: chord_range[1]]

            tracks[(idx, program, is_drum)] = grouped_score

        self.pprint("Grouping time: ", time.time() - start)
        # Now we have a dict of arrays for each track, grouped by bar, let's do the chord inference
        start = time.time()
        chords = fast_chord_inference(tracks, chord_durations)
        self.pprint("Chord inference time: ", time.time() - start)
        return chords, tracks, track_keys, tempo

    def write(self, chords, tracks, track_keys, tempo):
        """
        Write a score from the parsed data.
        You probably don't need to use this method directly, use MidiScore.write instead

        Parameters
        ----------
        chords : list
            List of chords with the time signature of each bar

        tracks : dict
            Dict of arrays with the notes for each track

        track_keys : list
            List of tuples with the track keys (program, is_drum, voice)

        tempo : float
            Tempo of the score


        Returns
        -------
        score : symusic.Score object
            The score object on which you can call dump_midi to save it to a file

        """
        score = Score(ttype=self.tpq)
        score.tempos.append(Tempo(0, tempo))
        current_ts = None
        for chord in chords:
            ts = TimeSignature(chord[7], chord[5], chord[6])
            candidate_ts = (ts.numerator, ts.denominator)
            if current_ts != candidate_ts:
                score.time_signatures.append(ts)
                current_ts = (
                    (ts.numerator, ts.denominator) if current_ts is None else current_ts
                )

        for track_key in track_keys:
            idx, program, is_drum = track_key
            track = Track(program=program, is_drum=is_drum)
            track.notes = tracks[(idx, program, is_drum)]
            score.tracks.append(track)
        return score

    def _get_bars(self, score, force_ts=None):
        """
        Get the bars of a score
        Parameters
        ----------
        score : symusic.Score object
            The score to get the bars from

        force_ts : tuple or None (Default value = None)
            Tuple with the time signature to force for the whole file. If None, use the time signature of the midi file

        Returns
        -------
        bars : list
            List of tuples with the start and end time of each bar in ticks
        chord_durations: list
            List of tuples with the time signature of each bar

        """
        # Calculate bars
        if force_ts is not None:
            time_signatures = [TimeSignature(0, force_ts[0], force_ts[1])]
        else:
            time_signatures = score.time_signatures
            time_signatures = sorted(time_signatures, key=lambda x: x.time)
            if len(time_signatures) == 0:
                time_signatures = [TimeSignature(0, 4, 4)]

        ticks_per_quarter = score.ticks_per_quarter
        chord_durations = []
        time = 0
        bars = []
        max_note_onset = max([n.time for t in score.tracks for n in t.notes])
        while time <= max_note_onset:
            start_time = time
            end_time, num, den = self._get_end_of_bar(
                time, time_signatures, ticks_per_quarter
            )
            chord_durations.append((num, den))
            bars.append((start_time, end_time))
            time = end_time
        return bars, chord_durations

    def _get_bar_nb_array_optimized(self, note_times, bars):
        """
        Get the bar number for each note in an optimized way
        Parameters
        ----------
        note_times : dict
            Dictionary with time key containing the note times

        bars : list
            List of tuples with the start and end time of each bar in ticks

        Returns
        -------
        bar_indices : np.array
            Array with the bar index for each note

        """
        # Convert bars to a NumPy array for efficient computation
        bars_array = np.array(bars)
        start_times = bars_array[:, 0]
        end_times = bars_array[:, 1]

        # Use broadcasting to find which bar each note belongs to
        # Note times are broadcast across the start and end times of bars to create a 2D boolean array
        in_bar = (note_times["time"][:, None] >= start_times) & (
                note_times["time"][:, None] < end_times
        )

        # Get the bar indices for each note. The result is a 2D array where each row has a single True value
        bar_indices = np.argmax(in_bar, axis=1)

        # Handle notes that do not fall within any bar by setting their bar index to a special value (e.g., -1)
        # This step is optional and depends on how you want to handle such cases
        outside_bars = ~in_bar.any(axis=1)
        bar_indices[outside_bars] = -1

        return bar_indices

    def _get_end_of_bar(self, bar_start, time_signatures, ticks_per_quarter):
        """
        Get the end of a bar given the start of the bar

        Parameters
        ----------
        bar_start : int
            The start of the bar in ticks

        time_signatures : list
            List of time signatures for the score

        ticks_per_quarter : int
            Ticks per quarter note for the midi file


        Returns
        -------
        end: int
            The end of the bar in ticks
        num: int
            The numerator of the time signature for the current bar
        den: int
            The denominator of the time signature for the current bar

        """
        # Find time signature
        # We use a tolerance of one quarter note to find the time signature
        res = [
            ts
            for ts in time_signatures
            if bar_start >= (ts.time - ticks_per_quarter // 2)
        ]
        time_signature = res[-1] if len(res) > 0 else TimeSignature(0, 4, 4)

        num = time_signature.numerator
        den = time_signature.denominator
        duration = 4 * frac(num, den)
        return int(bar_start + duration * ticks_per_quarter), num, den

    def _groupby_bar(self, score_soa, bars):
        """
        Group SOA by bar

        Parameters
        ----------
        score_soa : dict
            Dict of arrays with the notes for each track

        bars : list
            List of tuples with the start and end time of each bar in ticks

        Returns
        -------
        result : list
            List of dicts with the grouped notes for each bar

        """
        # Group SOA by bar
        bar_nb_array = score_soa["bar"]
        # Time is relative to the start of the bar in our notation
        del score_soa["bar"]
        result = [
            {k: score_soa[k][bar_nb_array == bar_nb] for k in score_soa.keys()}
            for bar_nb in range(len(bars))
        ]
        return result

    def pprint(self, *text):
        """
        Print debug information if verbose is True

        Parameters
        ----------
        *text :
            Text to print

        """
        if self.verbose:
            print(*text)

    def _preload_score(self, midi_file):
        """
        Quantize the score to the desired tpq and min_dur

        Parameters
        ----------
        midi_file : str
            Path to the midi file to parse


        Returns
        -------
        score_quantized : symusic.Score object
            The quantized score object

        """
        score = symusic.Score(midi_file, ttype="tick")
        score_quantized = score.resample(tpq=self.tpq, min_dur=self.min_dur)
        return score_quantized




import pytest
import numpy as np
from unittest.mock import patch, MagicMock
from maidi import MidiScore
from maidi.integrations.api import MusicLangAPI
from maidi import instrument
import os




@pytest.fixture
def setup_score():
    instruments = [instrument.PIANO]
    score = MidiScore.from_empty(instruments=instruments, nb_bars=8, ts=(4, 4), tempo=120)
    return score

MUSICLANG_API_KEY = "testapikey"


@pytest.fixture
def setup_api():
    return MusicLangAPI(api_key=MUSICLANG_API_KEY, verbose=True)

def test_init_api_with_env(setup_score):
    # Set API key in environment variable
    os.environ['MUSICLANG_API_KEY'] = MUSICLANG_API_KEY

    api = MusicLangAPI()

    os.environ.pop('MUSICLANG_API_KEY')
    with pytest.raises(ValueError) as e:
        api = MusicLangAPI()
    assert "API key missing" in str(e.value)


def test_get_subchords_and_subtags(setup_api):
    api = setup_api

    chords = [(0, 4, 'm', ''), (0, 4, 'm', ''), (1, 4, 'm', ''), (1, 4, 'm', '')]
    tags = [[[], [], [], []]]  # 1 track, 4 bars
    subchords, subtags = api._get_subchords_and_subtags(chords, tags, 1, 3, 1)
    assert subchords == [None, (0, 4, 'm', ''), (1, 4, 'm', '')]
    assert subtags == [[[], [], []]]



def test_check_tags_exists(setup_api):

    api = setup_api
    tags = [[['FAKE TAG']]]

    with pytest.raises(ValueError) as e:
        api._check_tags_exists(tags)
    assert "Unexisting tags have been used : {'FAKE TAG'}" in str(e.value)



def test_check_chord_exists_ok(setup_api):
    api = setup_api
    chords = [(0, 4, 'm', '6(sus4)')]

    api._check_chords_exists(chords)


def test_check_chord_exists_wrong_mode(setup_api):
    api = setup_api
    chords = [(0, 4, 'lydian', '')]

    with pytest.raises(ValueError) as e:
        api._check_chords_exists(chords)
    assert "Scale mode is wrong" in str(e.value)



def test_check_chord_degree_wrong(setup_api):
    api = setup_api
    chords = [(-1, 4, 'm', '')]

    with pytest.raises(ValueError) as e:
        api._check_chords_exists(chords)
    assert "Chord degree is wrong" in str(e.value)

def test_check_tonality_degree_wrong(setup_api):
    api = setup_api
    chords = [(0, -1, 'm', '')]

    with pytest.raises(ValueError) as e:
        api._check_chords_exists(chords)
    assert "Tonality degree is wrong" in str(e.value)

def test_check_extension_wrong(setup_api):
    api = setup_api
    chords = [(0, 0, 'm', 'wrong')]

    with pytest.raises(ValueError) as e:
        api._check_chords_exists(chords)
    assert "Wrong extension" in str(e.value)


@patch.object(MusicLangAPI, '_call_polling_api')
def test_poll_api_raise_error_when_failed(mock_call_polling_api, setup_api):

    def side_effect(*args, **kwargs):
        return {"status": "FAILED"}

    mock_call_polling_api.side_effect = side_effect

    api = setup_api
    with pytest.raises(RuntimeError) as e:
        api.poll_api("test_id")
    assert "Task failed" in str(e.value)


@patch.object(MusicLangAPI, '_call_polling_api')
def test_poll_api_returns_none_when_pending(mock_call_polling_api, setup_api):

    def side_effect(*args, **kwargs):
        return {"status": "PENDING"}

    mock_call_polling_api.side_effect = side_effect

    api = setup_api
    result = api.poll_api("test_id")
    assert result is None


from maidi.analysis.parse_notes import parse_pitch
from maidi.analysis import TagsProvider

INTERESTING_NOTES = [
    ("s", 1),  # The second note of the scale (9)
    ("s", 3), # The fourth note of the scale (11)
    ("s", 5), # The sixth note of the scale (13)
    ("h", 1), # The second note of the chromatic scale (b9)
    ("h", 2), # The third note of the chromatic scale (9)
    ("h", 3), # The fourth note of the chromatic scale (#9)
    ("h", 4), # The fifth note of the chromatic scale (natural 3)
    ("h", 6), # The seventh note of the chromatic scale (b5)
    ("h", 8), # The ninth note of the chromatic scale (b6)
    ("h", 9), # The tenth note of the chromatic scale (natural 6)
    ("h", 10), # The eleventh note of the chromatic scale (b7)
    ("h", 11), # The twelfth note of the chromatic scale (natural 7)
]


def get_tags_names():
    """ """
    return [
        "CONTROL_SPECIAL_NOTE__" + note + str(idx) for note, idx in INTERESTING_NOTES
    ]


class SpecialNotesTagsProvider(TagsProvider):
    ALL_TAGS = get_tags_names()

    def get_tags(self, track_bar, chord, score):
        return get_special_notes_tags(track_bar, chord)

def _get_special_notes_tags(track_bar, chord):
    """Create histogram of (type, idx) weighted per duration

    Parameters
    ----------
    track_bar :
        param chord:
    chord :
        

    Returns
    -------

    """
    notes = []
    for idx_note in range(len(track_bar["time"])):
        pitch = track_bar["pitch"][idx_note]
        velocity = track_bar["velocity"][idx_note]
        time = track_bar["time"][idx_note]
        duration = track_bar["duration"][idx_note]
        type, idx, oct = parse_pitch(pitch, chord, False, is_absolute=False)
        notes.append((type, idx, oct, velocity, time, duration))

    # Create histogram of (type, idx) weighted per duration
    special_notes = {}
    for note in notes:
        type, idx, oct, velocity, time, duration = note
        if (type, idx) not in special_notes:
            special_notes[(type, idx)] = 0
        special_notes[(type, idx)] += duration

    # Normalize
    total_duration = sum([duration for duration in special_notes.values()])
    for key in special_notes:
        special_notes[key] /= total_duration

    # Find top 3, find if interesting notes in this top, if so returns it
    top_notes = sorted(special_notes.items(), key=lambda x: x[1], reverse=True)[:3]
    interesting_notes = []
    for note, weight in top_notes:
        if note in INTERESTING_NOTES:
            interesting_notes.append(note)
    return interesting_notes


def get_special_notes_tags(track_bar, chord):
    """

    Parameters
    ----------
    track_bar :
        
    chord :
        

    Returns
    -------

    """
    special_notes = _get_special_notes_tags(track_bar, chord)
    return ["CONTROL_SPECIAL_NOTE__" + note + str(idx) for note, idx in special_notes]



from .density import get_density_tags, DensityTagsProvider
from .polyphony import get_min_max_polyphony_tags, get_min_polyphony_tags, get_max_polyphony_tags
from .register import get_min_max_register_tags, get_min_max_registers
from .special_notes import get_special_notes_tags



CONTROL_PROVIDERS = [DensityTagsProvider]


import pytest
import numpy as np
from unittest.mock import patch
from maidi import MidiScore
from maidi.integrations.api import MusicLangAPI

API_URL = "https://api.mockmusiclang.io"
MUSICLANG_API_KEY = "testapikey"

@pytest.fixture
def setup_api():
    return MusicLangAPI(api_key=MUSICLANG_API_KEY, api_url=API_URL, verbose=True)

@pytest.fixture
def setup_score():
    instruments = ['piano']
    return MidiScore.from_empty(instruments=instruments, nb_bars=4, ts=(4, 4), tempo=120)

def test_initialization(setup_api):
    api = setup_api
    assert api.api_url == API_URL
    assert api.api_key == MUSICLANG_API_KEY
    assert api.verbose is True

def test_invalid_model(setup_api, setup_score):
    api = setup_api
    score = setup_score
    mask = np.ones((1, 4))
    with pytest.raises(ValueError) as excinfo:
        api.predict(score, mask, model="invalid_model")
    assert "Model invalid_model not existing" in str(excinfo.value)

def test_invalid_temperature(setup_api, setup_score):
    api = setup_api
    score = setup_score
    mask = np.ones((1, 4))
    with pytest.raises(ValueError) as excinfo:
        api.predict(score, mask, temperature=2.5)
    assert "Temperature must be lower than 2.0" in str(excinfo.value)

def test_invalid_mask_dimension(setup_api, setup_score):
    api = setup_api
    score = setup_score
    mask = np.ones((2, 4))  # Invalid mask shape
    with pytest.raises(ValueError) as excinfo:
        api.predict(score, mask)
    assert "Wrong number of tracks in the mask" in str(excinfo.value)

    mask = np.ones((1, 5))  # Invalid mask shape
    with pytest.raises(ValueError) as excinfo:
        api.predict(score, mask)
    assert "Wrong number of bars in the mask" in str(excinfo.value)

    mask = np.zeros((1, 4))  # Mask should not have only 0
    with pytest.raises(ValueError) as excinfo:
        api.predict(score, mask)
    assert "Mask must have at least one 1" in str(excinfo.value)

def test_invalid_chords(setup_api, setup_score):
    api = setup_api
    score = setup_score
    mask = np.ones((1, 4))
    invalid_chords = [(8, 0, 'M', '')]  # Invalid chord degree
    with pytest.raises(ValueError) as excinfo:
        api.predict(score, mask, chords=invalid_chords)
    assert "Chord degree is wrong" in str(excinfo.value)

def predict_side_effect(score, *args, **kwargs):
    return score  # Return the input segment as the result

@patch.object(MusicLangAPI, '_predict_with_api')
def test_predict_call(mock_predict_with_api, setup_api, setup_score):
    mock_predict_with_api.side_effect = predict_side_effect

    api = setup_api
    score = setup_score
    mask = np.ones((1, 4))
    api.predict(score, mask, model="control_masking_large")
    assert mock_predict_with_api.called

@patch.object(MusicLangAPI, '_predict_with_api')
def test_predict_with_tags(mock_predict_with_api, setup_api, setup_score):
    mock_predict_with_api.side_effect = predict_side_effect
    api = setup_api
    score = setup_score
    mask = np.ones((1, 4))
    tags = [[['CONTROL_MIN_POLYPHONY__1'] for _ in range(4)] for _ in range(1)]
    api.predict(score, mask, tags=tags)
    assert mock_predict_with_api.called

@patch.object(MusicLangAPI, '_predict_with_api')
def test_predict_with_chords(mock_predict_with_api, setup_api, setup_score):
    mock_predict_with_api.side_effect = predict_side_effect
    api = setup_api
    score = setup_score
    mask = np.ones((1, 4))
    chords = [(0, 0, 'M', '') for _ in range(4)]
    api.predict(score, mask, chords=chords)
    assert mock_predict_with_api.called

@patch.object(MusicLangAPI, '_predict_with_api')
def test_sanitize_score_return_good_track_nb(mock_predict_with_api, setup_api, setup_score):
    def predict_side_effect(score, *args, **kwargs):
        return score.concatenate(score, axis=0)
    mock_predict_with_api.side_effect = predict_side_effect

    api = setup_api
    score = setup_score
    mask = score.get_mask()
    mask[:, :] = 1
    result = api.predict(score, mask, model="control_masking_large")
    assert result.nb_tracks == score.nb_tracks


@patch.object(MusicLangAPI, '_predict_with_api')
def test_sanitize_score_return_good_bar_nb(mock_predict_with_api, setup_api, setup_score):
    def predict_side_effect(score, *args, **kwargs):
        return score.concatenate(score, axis=1)
    mock_predict_with_api.side_effect = predict_side_effect

    api = setup_api
    score = setup_score
    mask = score.get_mask()
    mask[:, :] = 1
    result = api.predict(score, mask, model="control_masking_large")
    assert result.nb_bars == score.nb_bars


@patch.object(MusicLangAPI, '_predict_with_api')
def test_predict_wrong_number_of_chords(mock_predict_with_api, setup_api, setup_score):
    def predict_side_effect(score, *args, **kwargs):
        return score.concatenate(score, axis=1)
    mock_predict_with_api.side_effect = predict_side_effect

    api = setup_api
    score = setup_score
    chords = [(0, 0, 'M', '') for _ in range(setup_score.nb_bars + 2)] # Wrong number of chords
    mask = score.get_mask()
    mask[:, :] = 1
    with pytest.raises(ValueError) as excinfo:
        api.predict(score, mask, chords=chords)
    assert "Wrong number of chords" in str(excinfo.value)


@patch.object(MusicLangAPI, '_predict_with_api')
def test_predict_too_high_context(mock_predict_with_api, setup_api, setup_score):
    def predict_side_effect(score, *args, **kwargs):
        return score.concatenate(score, axis=1)
    mock_predict_with_api.side_effect = predict_side_effect

    api = setup_api
    score = MidiScore.from_empty(instruments=['piano'], nb_bars=MusicLangAPI.MAX_CONTEXT + 4, ts=(4, 4), tempo=120)
    mask = score.get_mask()
    mask[:, :] = 1
    with pytest.raises(ValueError) as excinfo:
        api.predict(score, mask)
    assert "The prompted score must be less than" in str(excinfo.value)

@patch.object(MusicLangAPI, '_predict_with_api')
def test_predict_polling_value_negative_error(mock_predict_with_api, setup_api, setup_score):
    def predict_side_effect(score, *args, **kwargs):
        return score.concatenate(score, axis=1)
    mock_predict_with_api.side_effect = predict_side_effect

    api = setup_api
    score = setup_score
    mask = score.get_mask()
    mask[:, :] = 1
    with pytest.raises(ValueError) as excinfo:
        api.predict(score, mask, polling_interval=-1)
    assert "Polling interval must be > 0" in str(excinfo.value)

@patch.object(MusicLangAPI, '_call_predict_api')
@patch.object(MusicLangAPI, 'from_task_id')
def test_predict_with_api(mock_from_task_id, mock_call_predict_api, setup_api, setup_score):
    api = setup_api
    score = setup_score
    mask = score.get_mask()
    mask[:, :] = 1
    api._predict_with_api(score, mask, model="control_masking_large")
    assert mock_call_predict_api.called
    assert mock_from_task_id.called



@patch.object(MusicLangAPI, '_call_predict_api')
@patch.object(MusicLangAPI, 'from_task_id')
def test_predict_with_api_async(mock_from_task_id, mock_call_predict_api, setup_api, setup_score):
    api = setup_api
    score = setup_score
    mask = score.get_mask()
    mask[:, :] = 1


    api._predict_with_api(score, mask, async_mode=True)
    assert mock_call_predict_api.called
    assert not mock_from_task_id.called


class Tag:

    def __init__(self, root: str, value: str):
        self.root = root
        self.value = value

    def to_tag(self):
        return f"{self.root}__{self.value}"


class TagsProvider:
    ALL_TAGS = None

    def get_tags_list(self):
        if self.ALL_TAGS is None:
            raise ValueError("ALL_TAGS is not defined for the provider class, it should be a list of all tags")

        return self.ALL_TAGS

    def get_tags(self, track_bar, bar, score):
        raise NotImplementedError


    def get_pitches(self, track_bar, bar, score):
        pitches = track_bar["pitch"]
        return list(pitches)

    def get_start_end_notes(self, track_bar, bar, score):
        from maidi.utils.bar_helpers import bar_to_bar_duration_in_quarters
        times = track_bar["time"]
        durations = track_bar["duration"]
        ends = times + durations
        notes = list(zip(times, ends))
        return notes

class ScoreTagger:

    def __init__(self, tags_providers):
        self.tags_providers = tags_providers

    def remove_fake_notes(self, track_bar):
        """
        Remove notes that have velocity < 2

        """
        from maidi import MidiScore
        index = track_bar["velocity"] >= MidiScore.MINIMUM_VELOCITY
        new_track_bar = {}
        for key in track_bar.keys():
            new_track_bar[key] = track_bar[key][index]
        return new_track_bar


    @classmethod
    def get_base_tagger(cls):
        from maidi.analysis import tags_providers
        return ScoreTagger(
            [
                tags_providers.DensityTagsProvider(),
                tags_providers.MinMaxRegisterTagsProvider(),
                tags_providers.SpecialNotesTagsProvider(),
                tags_providers.MinMaxPolyphonyTagsProvider(),
            ]
        )

    def tag_score(self, score):
        tags = []

        for idx_track, track_key in enumerate(score.track_keys):
            track_tags = []
            for idx_bar in range(score.nb_bars):
                bar_tags = []
                track_bar = score.get_track_bar(idx_track, idx_bar)
                bar = score.bars[idx_bar]
                for provider in self.tags_providers:
                    track_bar = self.remove_fake_notes(track_bar)
                    bar_tags += provider.get_tags(track_bar, bar, score)
                track_tags.append(bar_tags)
            tags.append(track_tags)

        from maidi import TagManager
        return TagManager(tags)




from maidi import midi_library


def test_midi_library_list():
    files = midi_library.get_list_of_midi_files()
    assert len(files) > 0


def test_midi_library_get_works():
    files = midi_library.get_list_of_midi_files()
    file = midi_library.get_midi_file(files[0])
    assert file is not None



from maidi.analysis.control_features import get_min_max_registers


def test_get_min_max_registers():
    """ """
    assert get_min_max_registers([30, 40, 50]) == ["lowest", "contrabass"]
    assert get_min_max_registers([80, 90, 100]) == ["sopranissimo", "highest"]
    assert get_min_max_registers([45, 55, 65]) == ["baritone", "baritone"]
    assert get_min_max_registers([53, 74]) == ["contralto", "contralto"]
    assert get_min_max_registers([41, 42, 43]) == ["bass", "bass"]
    assert get_min_max_registers([0, 1, 2]) == ["lowest", "lowest"]
    assert get_min_max_registers([60, 130]) == ["soprano", "highest"]
    assert get_min_max_registers([36, 57]) == ["contrabass", "contrabass"]


def test_get_min_max_registers_empty():
    """ """
    assert get_min_max_registers([]) == []



from maidi.analysis.control_features.polyphony import (
    get_min_polyphony,
    get_max_polyphony,
)  # Assume your functions are in 'your_module'


def test_get_max_polyphony_empty():
    """ """
    assert get_max_polyphony([]) == 0, "Max polyphony of empty list should be 0"


def test_get_min_polyphony_empty():
    """ """
    assert get_min_polyphony([]) == 0, "Min polyphony of empty list should be 0"


def test_get_max_polyphony_no_overlap():
    """ """
    notes = [(1, 2), (3, 4), (5, 6)]
    assert (
        get_max_polyphony(notes) == 1
    ), "Max polyphony with no overlapping notes should be 1"


def test_get_min_polyphony_no_overlap():
    """ """
    notes = [(1, 2), (3, 4), (5, 6)]
    assert (
        get_min_polyphony(notes) == 1
    ), "Min polyphony with no overlapping notes should be 1"


def test_get_max_polyphony_full_overlap():
    """ """
    notes = [(1, 3), (1, 3), (1, 3)]
    assert (
        get_max_polyphony(notes) == 3
    ), "Max polyphony with fully overlapping notes should be 3"


def test_get_min_polyphony_full_overlap():
    """ """
    notes = [(1, 3), (1, 3), (1, 3)]
    assert (
        get_min_polyphony(notes) == 3
    ), "Min polyphony with fully overlapping notes should be 3"


def test_get_min_polyphony_full_overlap_2():
    """ """
    notes = [(1, 3), (1, 3), (1, 3), (4, 5), (4, 5)]
    assert (
        get_min_polyphony(notes) == 2
    ), "Min polyphony with fully overlapping notes should be 2"


def test_get_max_polyphony_partial_overlap():
    """ """
    notes = [(1, 4), (2, 5), (3, 6)]
    assert (
        get_max_polyphony(notes) == 3
    ), "Max polyphony with partial overlaps should be 3"


def test_get_min_polyphony_partial_overlap():
    """ """
    notes = [(1, 4), (2, 5), (3, 6)]
    assert (
        get_min_polyphony(notes) == 1
    ), "Min polyphony with partial overlaps should be 1 when first note starts"


def test_get_max_polyphony_same_time_start_end():
    """ """
    notes = [(1, 5), (2, 2), (3, 6)]
    assert (
        get_max_polyphony(notes) == 2
    ), "Max polyphony should handle notes that start and end at the same time"


def test_get_min_polyphony_same_time_start_end():
    """ """
    notes = [(1, 5), (2, 2), (3, 6)]
    assert (
        get_min_polyphony(notes) == 1
    ), "Min polyphony should handle notes that start and end at the same time"


def test_get_max_polyphony_complex():
    """ """
    notes = [(1, 5), (2, 3), (4, 7), (6, 8)]
    assert (
        get_max_polyphony(notes) == 2
    ), "Max polyphony should correctly handle complex overlapping"


def test_get_min_polyphony_complex():
    """ """
    notes = [(1, 5), (2, 3), (4, 7), (6, 8)]
    assert (
        get_min_polyphony(notes) == 1
    ), "Min polyphony should correctly handle complex overlapping"


def test_get_max_polyphony_starts_when_other_ends():
    """ """
    notes = [(0, 1), (1, 2)]
    assert (
        get_max_polyphony(notes) == 1
    ), "Max polyphony should handle notes that start when another ends"


def test_min_polyphony_not_zero():
    """ """
    notes = [(0, 1), (2, 3), (4, 5)]
    assert (
        get_min_polyphony(notes) == 1
    ), "Min polyphony should not be zero when silence between two notes"



I = 0
II = 1
III = 2
IV = 3
V = 4
VI = 5
VII = 6

C = 0
Db = 1
Cs = 1
D = 2
Eb = 3
Ds = 3
E = 4
Es = 5
Fb = 4
F = 5
Fs = 6
Gb = 6
G = 7
Gs = 8
Ab = 8
A = 9
As = 10
Bb = 10
B = 11
Bs = 0

minor = 'm'
major = 'M'

_root_position = ''
_triad = ''
_6 = '6'
_64 = '64'
_7 = '7'
_65 = '65'
_43 = '43'
_2 = '2'
_sus2 = '(sus2)'
_sus4 = '(sus4)'
ALL_EXTENSIONS = set([_root_position, _6, _64, _7, _65, _43, _2])
SPECIAL_EXTENSIONS = set([_sus4, _sus2])


def is_valid_extension(extension):
    corr_extension = extension.replace(_sus2, '').replace(_sus4, '')
    if extension in SPECIAL_EXTENSIONS:
        return True

    for ext in SPECIAL_EXTENSIONS:
        if extension.startswith(ext):
            return False

    return corr_extension in ALL_EXTENSIONS






from copy import deepcopy


class TagManager:

    def __init__(self, tags):
        """
        Create a tag manager

        Parameters
        ----------
        tags : list[list[list[str]]]
            The tags
        """
        if isinstance(tags, TagManager):
            self.tags = deepcopy(tags.tags)
        else:
            self.tags = deepcopy(tags)

    def add_to_tag(self, tag_category, value):
        return tag_category.add_value(self, value)

    def replace_all_with_tag(self, tag_category, new_tag):
        return tag_category.replace_all_with_tag(self, new_tag)

    def replace_tag(self, tag_to_change, new_tag):
        from maidi import TagManager
        from copy import deepcopy
        new_array = deepcopy(self.tags)
        for idx_track, track in enumerate(new_array):
            for idx_bar, bar in enumerate(track):
                for idx_value, tag in enumerate(bar):
                    if tag == tag_to_change:
                        new_array[idx_track][idx_bar][idx_value] = new_tag

        return TagManager(new_array)


    @classmethod
    def empty_from_score(cls, score):
        """
        Create an empty tag manager from a score

        Parameters
        ----------
        score : MidiScore
            The score to create the tag manager from

        Returns
        -------
        tag_manager : TagManager
            The tag manager
        """
        shape = score.shape
        tags = [[[] for _ in range(shape[1])] for _ in range(shape[0])]
        return cls(tags)

    @classmethod
    def from_chord_manager(cls, chord_manager, nb_tracks):
        """
        Create a tag manager from a score and a chord manager

        Parameters
        ----------
        chord_manager : ChordManager
            The chord manager in which to extract tags associated to chord progression
        nb_tracks : int
            The number of tracks
        """
        tags = [[[] for _ in range(len(chord_manager))] for _ in range(nb_tracks)]
        for idx_track in range(nb_tracks):
            for idx_bar in range(len(chord_manager)):
                tags[idx_track][idx_bar] = chord_manager.chords[idx_bar][4]
        return cls(tags)



    def add_tag(self, tag):
        """
        Add a tag

        Parameters
        ----------
        tag : str
            The tag to add

        Returns
        -------
        None
        """
        return self.add_tags([tag])


    def __repr__(self):
        repr = ""
        for idx_track, track in enumerate(self.tags):
            repr += f"Track {idx_track} \n"
            for bar_idx, bar in enumerate(track):
                repr += f" Bar {bar_idx} : " + str(bar) + "\n"
            repr += "\n"
        return repr


    def add_tags(self, tags):
        """
        Add tags

        Parameters
        ----------
        tags : list[str]
            The tags to add

        Returns
        -------
        None
        """
        for track in self.tags:
            for bar in track:
                bar.extend(tags)

    def add_tag_at_index(self, tag, track_index, bar_index):
        """
        Add a tag at a specific index

        Parameters
        ----------
        tag : str
            The tag to add
        track_index : int
            The track index
        bar_index : int
            The bar index

        Returns
        -------
        None
        """
        self.add_tags_at_index([tag], track_index, bar_index)

    def add_tags_at_index(self, tags, track_index, bar_index):
        """
        Add tags at a specific index

        Parameters
        ----------
        tags : list[str]
            The tags to add
        track_index : int
            The track index
        bar_index : int
            The bar index

        Returns
        -------
        None
        """
        self.tags[track_index][bar_index].extend(tags)


    def add_tag_for_track(self, tag, track_index):
        """
        Add a tag for a specific track

        Parameters
        ----------
        tag : str
            The tag to add
        track_index : int
            The track index
        """
        return self.add_tags_for_track([tag], track_index)


    def add_tags_for_track(self, tags, track_index):
        """
        Add tags for a specific track

        Parameters
        ----------
        tags : list[str]
            The tags to add
        track_index : int
        """
        for bar in self.tags[track_index]:
            bar.extend(tags)

    def add_tags_for_bar(self, tags, bar_index):
        """
        Add tags for a specific bar

        Parameters
        ----------
        tags : list[str]
            The tags to add
        bar_index : int
        """
        for track in self.tags:
            track[bar_index].extend(tags)

    def add_tag_for_bar(self, tag, bar_index):
        """
        Add a tag for a specific bar

        Parameters
        ----------
        tag : str
            The tag to add
        bar_index : int
        """
        return self.add_tags_for_bar([tag], bar_index)

    def remove_tags(self, tags):
        """
        Remove tags in the whole score

        Parameters
        ----------
        tags : list[str]
            The tags to remove

        """
        for track in self.tags:
            for bar in track:
                for tag in tags:
                    if tag in bar:
                        bar.remove(tag)

    def remove_tag(self, tag):
        """
        Remove a tag in the whole score

        Parameters
        ----------
        tag : str
            The tag to remove


        """
        for track in self.tags:
            for bar in track:
                if tag in bar:
                    bar.remove(tag)

    def remove_tag_for_track(self, tag, track_index):
        """
        Remove a tag for a specific track

        Parameters
        ----------
        tag : str
            The tag to remove
        track_index : int
            The track index

        """
        for bar in self.tags[track_index]:
            if tag in bar:
                bar.remove(tag)

    def remove_tags_for_track(self, tags, track_index):
        """
        Remove tags for a specific track

        Parameters
        ----------
        tags : list[str]
            The tags to remove
        track_index : int
            The track index

        """
        for idx, bar in enumerate(self.tags[track_index]):
            self.tags[track_index][idx] = [tag for tag in bar if tag not in tags]

    def remove_tags_at_index(self, tags, track_index, bar_index):
        """
        Remove tags at a specific index

        Parameters
        ----------
        tags : list[str]
            The tags to remove
        track_index : int
            The track index
        bar_index : int
            The bar index

        """
        self.tags[track_index][bar_index] = [tag for tag in self.tags[track_index][bar_index] if tag not in tags]

    def remove_tag_at_index(self, tag, track_index, bar_index):
        """
        Remove a tag at a specific index

        Parameters
        ----------
        tag : str
            The tag to remove
        track_index : int
            The track index
        bar_index : int
            The bar index

        """
        self.remove_tags_at_index([tag], track_index, bar_index)

    def remove_tags_for_bar(self, tags, bar_index):
        """
        Remove tags for a specific bar

        Parameters
        ----------
        tags : list[str]
            The tags to remove
        bar_index : int
            The bar index

        """
        for track in self.tags:
            track[bar_index] = [tag for tag in track[bar_index] if tag not in tags]

    def remove_tag_for_bar(self, tag, bar_index):
        """
        Remove a tag for a specific bar

        Parameters
        ----------
        tag : str
            The tag to remove
        bar_index : int
            The bar index

        """
        self.remove_tags_for_bar([tag], bar_index)


    def clear_tags(self):
        """
        Clear all tags
        """
        for track in self.tags:
            for bar in track:
                bar.clear()

    def clear_tags_for_track(self, track_index):
        """
        Clear all tags for a specific track

        Parameters
        ----------
        track_index : int
            The track index

        """
        for bar in self.tags[track_index]:
            bar.clear()

    def clear_tags_for_bar(self, bar_index):
        """
        Clear all tags for a specific bar

        Parameters
        ----------
        bar_index : int
            The bar index

        """
        for track in self.tags:
            track[bar_index].clear()

    def clear_tags_at_index(self, track_index, bar_index):
        """
        Clear all tags at a specific index

        Parameters
        ----------
        track_index : int
            The track index
        bar_index : int
            The bar index

        """
        self.tags[track_index][bar_index].clear()


    @property
    def shape(self):
        """
        Get the shape of the tag manager (should be the same as the score)
        """
        return len(self.tags), len(self.tags[0])

    def __len__(self):
        """
        Get the length of the tag manager (corresponding to the number of tracks)
        """
        return len(self.tags)

    def get_tags_at_index(self, track_index, bar_index):
        """
        Get the tags at a specific index

        Parameters
        ----------
        track_index : int
            The track index
        bar_index : int
            The bar index

        Returns
        -------
        list[str]
            The tags at the specified index

        """
        return self.tags[track_index][bar_index]

    def filter(self, function):
        """
        Filter the tags, keeping only the ones that match the function

        Parameters
        ----------
        function : function
            The function to filter the tags

        Returns
        -------
        TagManager
            The filtered tag manager

        """
        tags = [[list(filter(function, bar)) for bar in track] for track in self.tags]
        return TagManager(tags)

    def filter_remove(self, function):
        """
        Filter the tags, removing the ones that match the function

        Parameters
        ----------
        function : function
            The function to filter the tags

        Returns
        -------
        TagManager
            The filtered tag manager

        """
        tags = [[list(filter(lambda x: not function(x), bar)) for bar in track] for track in self.tags]
        return TagManager(tags)

    def __getitem__(self, item):
        """
        Get the tags for a specific item

        Parameters
        ----------
        item : int, slice or tuple
            The item to get

        Returns
        -------
        TagManager
            The sub tag manager

        """
        if isinstance(item, slice):
            return TagManager(self.tags[item])
        if isinstance(item, int):
            return TagManager([self.tags[item]])
        if isinstance(item, tuple):
            if len(item) != 2:
                raise ValueError("Item should be a tuple of length 2")
            first_dim, second_dim = item[0], item[1]
            tags_first_dim = self[first_dim].tags
            if isinstance(second_dim, slice):
                return TagManager([track[second_dim] for track in tags_first_dim])
            if isinstance(second_dim, int):
                return TagManager([[track[second_dim]] for track in tags_first_dim])


    def __setitem__(self, item, value):
        """
        Set the tags for a specific item.

        Parameters
        ----------
        item : int, slice, or tuple
            The item to set.
        value : TagManager, str, list[str], or list[list[str]]
            The value to set. Can be a TagManager with the same shape as the query,
            a single string (assign [value] to all cells of the query), a list of strings
            (assign value to the query), or a list of list of strings.

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If the value does not match the expected shape or type.
        """
        # If value is a list of list of list of strings, convert it to a TagManager
        if isinstance(value, list) and all(isinstance(v, list) and all(isinstance(s, list) for s in v) for v in value):
            value = TagManager(value)

        if isinstance(item, int):
            # Single track assignment
            if isinstance(value, TagManager):
                if len(value.tags) != 1 or len(value.tags[0]) != len(self.tags[item]):
                    raise ValueError(
                        "Shape mismatch: TagManager must have the same shape as the queried sub TagManager.")
                self.tags[item] = value.tags[0]
            elif isinstance(value, str):
                self.tags[item] = [[value] for _ in range(len(self.tags[item]))]
            elif isinstance(value, list):
                if all(isinstance(v, str) for v in value):
                    self.tags[item] = [value for _ in range(len(self.tags[item]))]
                elif all(isinstance(v, list) and all(isinstance(s, str) for s in v) for v in value):
                    if len(value) != len(self.tags[item]):
                        raise ValueError("Shape mismatch: list of lists must match the number of bars in the track.")
                    self.tags[item] = value
                else:
                    raise ValueError("Invalid value type: must be a list of strings or a list of list of strings.")
            elif value is None:
                self.tags[item] = [[] for _ in range(len(self.tags[item]))]
            else:
                raise ValueError(
                    "Invalid value type: must be a TagManager, string, list of strings, None, or list of list of strings.")
        elif isinstance(item, slice):
            # Slice assignment
            if isinstance(value, TagManager):
                if len(value.tags) != len(self.tags[item]):
                    raise ValueError(
                        "Shape mismatch: TagManager must have the same shape as the queried sub TagManager.")
                self.tags[item] = value.tags
            elif isinstance(value, str):
                for i in range(*item.indices(len(self.tags))):
                    self.tags[i] = [[value] for _ in range(len(self.tags[i]))]
            elif value is None:
                for i in range(*item.indices(len(self.tags))):
                    self.tags[i] = [[] for _ in range(len(self.tags[i]))]
            elif isinstance(value, list):
                if all(isinstance(v, str) for v in value):
                    for i in range(*item.indices(len(self.tags))):
                        self.tags[i] = [value for _ in range(len(self.tags[i]))]
                elif all(isinstance(v, list) and all(isinstance(s, str) for s in v) for v in value):
                    if len(value) != len(self.tags[item]):
                        raise ValueError("Shape mismatch: list of lists must match the number of tracks in the slice.")
                    self.tags[item] = value
                else:
                    raise ValueError("Invalid value type: must be a list of strings or a list of list of strings.")
            else:
                raise ValueError(
                    "Invalid value type: must be a TagManager, string, list of strings, or list of list of strings.")
        elif isinstance(item, tuple):
            if len(item) != 2:
                raise ValueError("Item should be a tuple of length 2")
            first_dim, second_dim = item[0], item[1]
            if isinstance(first_dim, int):
                if isinstance(second_dim, int):
                    # Single cell assignment
                    if isinstance(value, str):
                        self.tags[first_dim][second_dim] = [value]
                    elif isinstance(value, list) and all(isinstance(v, str) for v in value):
                        self.tags[first_dim][second_dim] = value
                    else:
                        raise ValueError("Invalid value type: must be a string or list of strings.")
                elif isinstance(second_dim, slice):
                    # Slice of bars in a single track
                    if isinstance(value, TagManager):
                        if len(value.tags) != 1 or len(value.tags[0]) != len(self.tags[first_dim][second_dim]):
                            raise ValueError(
                                "Shape mismatch: TagManager must have the same shape as the queried sub TagManager.")
                        self.tags[first_dim][second_dim] = value.tags[0]
                    elif isinstance(value, str):
                        for i in range(*second_dim.indices(len(self.tags[first_dim]))):
                            self.tags[first_dim][i] = [value]
                    elif isinstance(value, list):
                        if all(isinstance(v, str) for v in value):
                            for i in range(*second_dim.indices(len(self.tags[first_dim]))):
                                self.tags[first_dim][i] = value
                        elif all(isinstance(v, list) and all(isinstance(s, str) for s in v) for v in value):
                            if len(value) != len(self.tags[first_dim][second_dim]):
                                raise ValueError(
                                    "Shape mismatch: list of lists must match the number of bars in the slice.")
                            self.tags[first_dim][second_dim] = value
                        else:
                            raise ValueError(
                                "Invalid value type: must be a list of strings or a list of list of strings.")
                    elif value is None:
                        self.tags[first_dim][second_dim] = [[] for _ in self.tags[first_dim][second_dim]]
                    else:
                        raise ValueError(
                            "Invalid value type: must be a TagManager, string, None, list of strings, or list of list of strings.")
            elif isinstance(first_dim, slice):
                if isinstance(second_dim, int):
                    # Single bar across multiple tracks
                    if isinstance(value, TagManager):
                        if len(value.tags) != len(self.tags[first_dim]) or any(len(track) != 1 for track in value.tags):
                            raise ValueError(
                                "Shape mismatch: TagManager must have the same shape as the queried sub TagManager.")
                        for i, track in enumerate(self.tags[first_dim]):
                            track[second_dim] = value.tags[i][0]
                    elif isinstance(value, str):
                        for i in range(*first_dim.indices(len(self.tags))):
                            self.tags[i][second_dim] = [value]
                    elif value is None:
                        for i in range(*first_dim.indices(len(self.tags))):
                            self.tags[i][second_dim] = []
                    elif isinstance(value, list) and all(isinstance(v, str) for v in value):
                        for i in range(*first_dim.indices(len(self.tags))):
                            self.tags[i][second_dim] = value
                    else:
                        raise ValueError("Invalid value type: must be a TagManager, string, or list of strings.")
                elif isinstance(second_dim, slice):
                    # Slice of bars across multiple tracks
                    if isinstance(value, TagManager):
                        if len(value.tags) != len(self.tags[first_dim]) or any(
                                len(track) != len(self.tags[first_dim][0][second_dim]) for track in value.tags):
                            raise ValueError(
                                "Shape mismatch: TagManager must have the same shape as the queried sub TagManager.")
                        for i, track in enumerate(self.tags[first_dim]):
                            track[second_dim] = value.tags[i]
                    elif isinstance(value, str):
                        for i in range(*first_dim.indices(len(self.tags))):
                            for j in range(*second_dim.indices(len(self.tags[i]))):
                                self.tags[i][j] = [value]
                    elif value is None:
                        for i in range(*first_dim.indices(len(self.tags))):
                            for j in range(*second_dim.indices(len(self.tags[i]))):
                                self.tags[i][j] = []
                    elif isinstance(value, list):
                        if all(isinstance(v, str) for v in value):
                            for i in range(*first_dim.indices(len(self.tags))):
                                for j in range(*second_dim.indices(len(self.tags[i]))):
                                    self.tags[i][j] = value
                        elif all(isinstance(v, list) and all(isinstance(s, str) for s in v) for v in value):
                            if len(value) != len(self.tags[first_dim]) or any(
                                    len(v) != len(self.tags[first_dim][0][second_dim]) for v in value):
                                raise ValueError(
                                    "Shape mismatch: list of lists must match the number of tracks and bars in the slice.")
                            for i, track in enumerate(self.tags[first_dim]):
                                track[second_dim] = value[i]
                        else:
                            raise ValueError(
                                "Invalid value type: must be a list of strings or a list of list of strings.")
                    else:
                        raise ValueError(
                            "Invalid value type: must be a TagManager, string, list of strings, or list of list of strings.")
            else:
                raise ValueError("Invalid item type: must be an int, slice, or tuple of length 2.")
        else:
            raise ValueError("Invalid item type: must be an int, slice, or tuple of length 2.")